'''
 
  non-linear stationary solver

'''
import mfem.common.chypre as chypre
import petram.helper.block_matrix as bm
from petram.solver.solver_model import SolverInstance
import os
import numpy as np
from scipy.sparse import coo_matrix

from petram.model import Model
from petram.solver.solver_model import Solver
import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('NLSolver')


class NLSolver(Solver):
    can_delete = True
    has_2nd_panel = False

    @classmethod
    def fancy_menu_name(self):
        return 'Stationary(NL)'

    @classmethod
    def fancy_tree_name(self):
        return 'NLStationary'

    def attribute_set(self, v):
        super(NLSolver, self).attribute_set(v)
        v["nl_scheme"] = "Newton"
        v["nl_maxiter"] = 10
        v["nl_reltol"] = 0.001
        v["nl_abstol"] = 0.0
        v["nl_damping"] = 1.0
        v["nl_damping_min"] = 0.01
        v["nl_damping_fixed"] = False
        v["nl_tolthr"] = 0.01
        v["nl_stopcond"] = "error"
        v["nl_stopresidual"] = "all"
        v["nl_verbose"] = True
        v['dwc_name'] = ''
        v['use_dwc_nl'] = False
        v['dwc_nl_arg'] = ''

        return v

    def panel1_param(self):
        ret = [["dwc",   self.dwc_name,   0, {}],
               ["args.",   self.dwc_nl_arg,   0, {}]]
        value = [self.dwc_name, self.dwc_nl_arg]
        return [  # ["Initial value setting",   self.init_setting,  0, {},],
            ["physics model",   self.phys_model,  0, {}, ],
            ["Non-linear solver", None, 1, {
                "values": ["FixedPoint", "Newton"]}],
            ["Max iteration", self.nl_maxiter, 400, {}],
            ["NL rel. tol.", self.nl_reltol, 300, {}],
            ["NL inital damping", self.nl_damping, 300, {}],
            ["NL min damping", self.nl_damping_min, 300, {}],
            [None, not self.nl_damping_fixed,  3,
                {"text": "auto adjust damping"}],
            [None, [False, value], 27, [{'text': 'Use DWC (nl_start/nl_checkpoint/nl_end)'},
                                        {'elp': ret}]],
            [None, self.nl_verbose, 3, {
                "text": "verbose output for non-linear iteration"}],
            [None, self.init_only,  3, {"text": "initialize solution only"}],
            [None,
             self.clear_wdir,  3, {"text": "clear working directory"}],
            [None,
             self.assemble_real,  3, {"text": "convert to real matrix (complex prob.)"}],
            [None,
             self.save_parmesh,  3, {"text": "save parallel mesh"}],
            [None,
             self.use_profiler,  3, {"text": "use profiler"}],
            [None, self.skip_solve,  3, {"text": "skip linear solve"}],
            [None, self.load_sol,  3, {
                "text": "load sol file (linear solver is not called)"}],
            [None, self.sol_file,  0, None], ]

    def get_panel1_value(self):
        return (  # self.init_setting,
            self.phys_model,
            self.nl_scheme,
            self.nl_maxiter,
            self.nl_reltol,
            self.nl_damping,
            self.nl_damping_min,
            not self.nl_damping_fixed,
            [self.use_dwc_nl, [self.dwc_name, self.dwc_nl_arg, ]],
            self.nl_verbose,
            self.init_only,
            self.clear_wdir,
            self.assemble_real,
            self.save_parmesh,
            self.use_profiler,
            self.skip_solve,
            self.load_sol,
            self.sol_file,)

    def import_panel1_value(self, v):
        #self.init_setting = str(v[0])
        self.phys_model = str(v[0])
        self.nl_scheme = v[1]
        self.nl_maxiter = v[2]
        self.nl_reltol = v[3]
        self.nl_damping = v[4]
        self.nl_damping_min = v[5]
        self.nl_damping_fixed = not v[6]
        self.use_dwc_nl = v[7][0]
        self.dwc_name = v[7][1][0]
        self.dwc_nl_arg = v[7][1][1]

        self.nl_verbose = v[8]

        self.init_only = v[9]
        self.clear_wdir = v[10]
        self.assemble_real = v[11]
        self.save_parmesh = v[12]
        self.use_profiler = v[13]
        self.skip_solve = v[14]
        self.load_sol = v[15]
        self.sol_file = v[16]

    def get_editor_menus(self):
        return []

    def get_possible_child(self):
        choice = []
        try:
            from petram.solver.mumps_model import MUMPS
            choice.append(MUMPS)
        except ImportError:
            pass

        # try:
        #    from petram.solver.gmres_model import GMRES
        #    choice.append(GMRES)
        # except ImportError:
        #    pass

        try:
            from petram.solver.iterative_model import Iterative
            choice.append(Iterative)
        except ImportError:
            pass

        try:
            from petram.solver.strumpack_model import Strumpack
            choice.append(Strumpack)
        except ImportError:
            pass
        return choice

    def allocate_solver_instance(self, engine):
        if self.clear_wdir:
            engine.remove_solfiles()

        if self.nl_scheme == 'Newton':
            instance = NewtonSolver(
                self, engine) if self.instance is None else self.instance
        elif self.nl_scheme == 'FixedPoint':
            instance = FixedPointSolver(
                self, engine) if self.instance is None else self.instance
        else:
            assert False, "Unknown Nonlinear solver:" + self.nl_scheme

        return instance

    def get_matrix_weight(self, timestep_config):  # , timestep_weight):
        # this solver uses y, and grad(y)
        from petram.engine import max_matrix_num
        weight = [0]*max_matrix_num

        if timestep_config[0]:
            weight[0] = 1
            if self.nl_scheme == 'Newton':
                weight[max_matrix_num//2] = 1
        return weight

    @debug.use_profiler
    def run(self, engine, is_first=True, return_instance=False):
        dprint1("Entering run (is_first= ", is_first, ") ", self.fullpath())

        instance = self.allocate_solver_instance(engine)
        instance.set_verbose(self.nl_verbose)
        instance.set_tol(self.nl_reltol, self.nl_tolthr)
        instance.set_blk_mask()
        if return_instance:
            return instance

        instance.configure_probes(self.probe)
        engine.sol = engine.assembled_blocks[1][0]
        instance.sol = engine.sol

        if self.init_only:
            pass
        elif self.load_sol:
            if is_first:
                instance.assemble()
                is_first = False
            instance.load_sol(self.sol_file)
        else:
            instance.reset_count(self.nl_maxiter)
            instance.set_damping(self.nl_damping,
                                 minimum=self.nl_damping_min,
                                 fixed=self.nl_damping_fixed)
            dprint1("Starting non-linear iteration")

            if self.use_dwc_nl:
                engine.call_dwc(self.get_phys_range(),
                                method="nl_start",
                                callername=self.name(),
                                dwcname=self.dwc_name,
                                args=self.dwc_nl_arg)

            while not instance.done:
                dprint1("="*72)
                dprint1("NL iteration step=", instance.kiter)
                if is_first:
                    instance.assemble()
                    is_first = False
                else:
                    instance.assemble(update=True)

                update_operator = engine.check_block_matrix_changed(
                    instance.blk_mask)
                instance.solve(update_operator=update_operator)

                if not instance.done:
                    # we do this only if we are going into the next loop
                    # since the same is done in save_solution
                    instance.recover_solution(ksol=0)

            if self.use_dwc_nl:
                engine.call_dwc(self.get_phys_range(),
                                method="nl_end",
                                callername=self.name(),
                                dwcname=self.dwc_name,
                                args=self.dwc_nl_arg)

        instance.save_solution(ksol=0,
                               skip_mesh=False,
                               mesh_only=False,
                               save_parmesh=self.save_parmesh)
        engine.sol = instance.sol

        instance.save_probe()

        self.instance = instance

        dprint1(debug.format_memory_usage())
        return is_first


class NonlinearBaseSolver(SolverInstance):
    def __init__(self, gui, engine):
        SolverInstance.__init__(self, gui, engine)
        self.assembled = False
        self.linearsolver = None
        self._operator_set = False
        self._kiter = 0
        self._alpha = 1.0
        self._beta = 0.0
        self._done = False
        self._converged = False
        self._verbose = False
        self.debug_data = []
        self.debug_data2 = []
        self.damping_record = []
        self.residual_record = []

    @property
    def blocks(self):
        return self.engine.assembled_blocks

    @property
    def kiter(self):
        return self._kiter

    @property
    def done(self):
        return self._done

    @property
    def verbose(self):
        return self._verbose

    def set_verbose(self, verbose):
        self._verbose = verbose

    def set_tol(self, reltol, tolthr):
        self._reltol = reltol
        self._tolthr = tolthr
        dprint1("NL iteration tolelance/threshold: ",
                self._reltol, self._tolthr)

    def set_damping(self, damping, minimum=None, fixed=None):
        assert False, "Must be implemented in child"

    def compute_A(self, M, B, X, mask_M, mask_B):
        assert False, "Must be implemented in subclass"

    def compute_rhs(self, M, B, X):
        assert False, "Must be implemented in subclass"

    def assemble_rhs(self):
        assert False, "assemble_rhs should not be called"

    def reset_count(self, maxiter):
        self._kiter = 0
        self._maxiter = maxiter
        self._current_error = (np.infty, np.infty)
        self._done = False
        self._converged = False
        self.debug_data = []
        self.error_record = []
        self.damping_record = []

    def assemble(self, inplace=True, update=False):
        engine = self.engine
        phys_target = self.get_phys()
        phys_range = self.get_phys_range()

        # use get_phys to apply essential to all phys in solvestep
        dprint1("Asembling system matrix",
                [x.name() for x in phys_target],
                [x.name() for x in phys_range])

        if not update:
            engine.run_verify_setting(phys_target, self.gui)
        else:
            engine.set_update_flag('TimeDependent')

        M_updated = engine.run_assemble_mat(
            phys_target, phys_range, update=update)
        B_updated = engine.run_assemble_b(phys_target, update=update)

        engine.run_apply_essential(phys_target, phys_range, update=update)
        engine.run_fill_X_block(update=update)

        _blocks, M_changed = self.engine.run_assemble_blocks(self.compute_A,
                                                             self.compute_rhs,
                                                             inplace=inplace,
                                                             update=update,)

        #A, X, RHS, Ae, B, M, names = _blocks

        self.assembled = True
        return M_changed

    def do_solve(self, update_operator=True):
        update_operator = update_operator or not self._operator_set
        engine = self.engine

        # if not self.assembled:
        #    assert False, "assmeble must have been called"

        A, X, RHS, Ae, B, M, depvars = self.blocks

        mask = self.blk_mask
        engine.copy_block_mask(mask)

        depvars = [x for i, x in enumerate(depvars) if mask[0][i]]

        if update_operator:
            opr = self.adjust_operator(A, M, mask)   # add Jacobian for Newtown
            AA = engine.finalize_matrix(opr, mask, not self.phys_real,
                                        format=self.ls_type)

        BB = engine.finalize_rhs([RHS], A, X[0], mask, not self.phys_real,
                                 format=self.ls_type,
                                 use_residual=True)

        if self.linearsolver is None:
            linearsolver = self.allocate_linearsolver(
                self.gui.is_complex(), self. engine)
            self.linearsolver = linearsolver
        else:
            linearsolver = self.linearsolver

        linearsolver.skip_solve = self.gui.skip_solve

        if update_operator:
            linearsolver.SetOperator(AA,
                                     dist=engine.is_matrix_distributed,
                                     name=depvars)
            self._operator_set = True

        if linearsolver.is_iterative:
            XX = engine.finalize_x(X[0], RHS, mask, not self.phys_real,
                                   format=self.ls_type)
        else:
            XX = None

        solall = linearsolver.Mult(BB, x=XX, case_base=0)

        #linearsolver.SetOperator(AA, dist = engine.is_matrix_distributed)
        #solall = linearsolver.Mult(BB, case_base=0)

        if not self.phys_real and self.gui.assemble_real:
            solall = self.linearsolver_model.real_to_complex(solall, AA)

        return solall

    def update_x(self, solall):
        A, X, RHS, Ae, B, M, depvars = self.blocks
        mask = self.blk_mask

        A.reformat_central_mat(
            solall, 0, X[0], mask, alpha=self._alpha, beta=self._beta)
        self.sol = X[0]

        # store probe signal (use t=0.0 in std_solver)
        for p in self.probe:
            p.append_sol(X[0])

        self._kiter = self._kiter + 1

    def do_error_estimate(self):
        '''
        call do_solve without updating operator
        '''
        damping = self.damping
        self._alpha = 1.0
        self._beta = 0.0

        tmp = self.do_solve(update_operator=False)
        self.update_x(tmp)

        self.set_damping(damping)
        self._kiter = self._kiter - 1

    def call_dwc_nliteration(self):
        if self.gui.use_dwc_nl:
            stopit = self.engine.call_dwc(self.gui.get_phys_range(),
                                          method="nl_checkpoint",
                                          callername=self.gui.name(),
                                          dwcname=self.gui.dwc_name,
                                          args=self.gui.dwc_nl_arg,
                                          count=self.kiter-1,)
            return stopit
        else:
            return False

    def load_sol(self, solfile):
        from petram.mfem_config import use_parallel
        if use_parallel:
            from mpi4py import MPI
        else:
            from petram.helper.dummy_mpi import MPI
        myid = MPI.COMM_WORLD.rank

        if myid == 0:
            solall = np.load(solfile)
        else:
            solall = None

        A, X, RHS, Ae, B, M, depvars = self.blocks
        mask = self.blk_mask
        A.reformat_central_mat(solall, 0, X[0], mask)
        self.sol = X[0]

        # store probe signal (use t=0.0 in std_solver)
        for p in self.probe:
            p.append_sol(X[0])

        return True

    def copy_x(self, X):
        '''
        extract data from X (BlockMatrix)
        '''
        shape = X.shape

        xdata = []
        for i in range(shape[0]):
            for j in range(shape[1]):
                v = X[i, j]
                if isinstance(v, chypre.CHypreVec):
                    vec = v.toarray()
                elif isinstance(v, bm.ScipyCoo):
                    vec = v.toarray()
                else:
                    assert False, "not supported"
                xdata.append(vec.copy())
        return xdata

    def copyback_x(self, X, xdata):
        '''
        put data back to X (BlockMatrix)
        '''
        shape = X.shape

        for i in range(shape[0]):
            for j in range(shape[1]):
                v = X[i, j]
                if isinstance(v, chypre.CHypreVec):
                    if v.isComplex():
                        v.real.Assign(xdata[i].real)
                        v.imag.Assign(xdata[i].imag)
                    else:
                        v.real.Assign(xdata[i])
                elif isinstance(v, bm.ScipyCoo):
                    coo = coo_matrix(xdata[i])
                    v.data = coo.data
                    v.row = coo.row
                    v.col = coo.col
                else:
                    assert False, "not supported"

        return xdata


class NewtonSolver(NonlinearBaseSolver):
    def __init__(self, gui, engine):
        NonlinearBaseSolver.__init__(self, gui, engine)
        self.scheme_name = "newton"
        self.minimum_damping = 0.05
        self._err_before = 100.
        self._err_guidance = 100.
        self._err_inc_count = 0
        self._fixed_damping = False
        self._stall_counter = 0
        self.max_stall = 10

        self._dwidth = 0.15
        self.dwidth1 = 1 + self._dwidth
        self.dwidth2 = 1 - self._dwidth

    def reset_count(self, maxiter):
        NonlinearBaseSolver.reset_count(self, maxiter)
        self._err_before = 100.

    @property
    def damping(self):
        return self._alpha

    def set_damping(self, damping, minimum=None, fixed=None):
        self._alpha = min(damping, 1.0)
        self._beta = 1.0
        if minimum is not None:
            self.minimum_damping = minimum
        if fixed is not None:
            self._fixed_damping = fixed

    def reset_count(self, maxiter):
        NonlinearBaseSolver.reset_count(self, maxiter)

    def compute_A(self, M, B, X, mask_M, mask_B):
        '''
        return A and isAnew
        '''
        A = M[0]
        return A, np.any(mask_M[0])

    def compute_rhs(self, M, B, X):
        '''
        M[0] x = B
        '''
        return B

    def compute_err(self, sol, sol_ave_norm, Res):
        '''
        sol is list of dense array
        Res is BlockMatrix
        '''
        shape = Res.shape
        assert shape[1] == 1, "multilpe vectors are not supported"

        err = np.zeros(shape[0])
        length = np.zeros(shape[0])

        for i in range(shape[0]):
            for j in [0]:
                res = Res[i]
                x_vec = sol[i]
                if isinstance(res, chypre.CHypreVec):
                    res_vec = res.toarray()
                elif isinstance(res, bm.ScipyCoo):
                    res_vec = res.toarray()
                else:
                    assert False, "not supported"
                w = np.abs(x_vec)
                if sol_ave_norm[i] != 0:
                    thr = sol_ave_norm[i]*self._tolthr
                else:
                    thr = np.mean(sol_ave_norm)*self._tolthr
                w[w < thr] = thr
                err_est = np.abs(res_vec)

                err[i] = np.sum((err_est/w)**2)
                length[i] = np.prod(x_vec.shape)

        from petram.mfem_config import use_parallel
        if use_parallel:
            from mpi4py import MPI
            allgather = MPI.COMM_WORLD.allgather
            length = np.sum(allgather(length), 0)
            err = np.sum(allgather(err), 0)

        err = err/length
        self.debug_data.append(err)

        err_total = np.sqrt(np.sum(err))/np.sqrt(shape[0])

        return err_total

    def compute_residual(self, RHS):
        shape = RHS.shape
        w = np.array([0]*shape[0])
        w[0] = 1

        if self.kiter == 0:
            self._res_sq0 = RHS.average_norm(sq=True)
            return self._res_sq0
        if self.kiter == 1:
            self._res_sq1 = RHS.average_norm(sq=True)
            self._res_w = (self._res_sq0 + self._res_sq1)/2.0
            self._res_w = [w if w != 0 else np.mean(self._res_w)
                           for w in self._res_w]
            return self._res_sq1
        if self.kiter > 1:
            nsq = RHS.normsq()
            dprint1(np.sqrt(np.sum(w*nsq/self._res_w))/np.sqrt(shape[0]),
                    nsq)
            return np.sqrt(np.sum(w*nsq/self._res_w))/np.sqrt(shape[0])

    def assemble(self, inplace=True, update=False):
        from petram.engine import max_matrix_num

        if self.kiter == 0:
            self.engine.deactivate_matrix(max_matrix_num//2)
        else:
            self.engine.activate_matrix(max_matrix_num//2)

        M_changed = NonlinearBaseSolver.assemble(
            self, inplace=inplace, update=update)

        if self.kiter > 0:
            A, X, RHS, Ae, B, M, depvars = self.blocks
            self.engine.eliminateJac(M[max_matrix_num//2])

    def solve(self, update_operator=True):
        A, X, RHS, Ae, B, M, depvars = self.blocks

        if self.verbose:
            dprint1("Linear solve...step=", self.kiter)

        residual = self.compute_residual(B - M[0].dot(X[0]))
        if self.kiter == 2:
            self._resitual0 = residual
        if self.verbose:
            dprint1("Current residual..", residual)

        if self.kiter > 0:

            sol_ave_norm = X[0].average_norm()
            soldata = self.copy_x(X[0])

            self.do_error_estimate()
            err = self.compute_err(soldata, sol_ave_norm, X[0])
            self.copyback_x(X[0], soldata)

            if self.verbose:
                dprint1("estimated error, err_before, err_guidance, err_inc_count,  damping)",
                        err, self._err_before, self._err_guidance, self._err_inc_count, self.damping)
            if err < self._reltol:
                self._converged = True
                self._done = True
            if abs((err - self._err_before)/err) < 1e-3:
                self._stall_counter = self._stall_counter + 1
            else:
                self._stall_counter = 0
            if self._stall_counter > self.max_stall:
                dprint1("no convergence (stall)")
                self._done = True

            if self.kiter == 1:
                self._err_before = err
                self._err_guidance = err
            elif self._fixed_damping:
                pass

            elif (err > self._err_guidance*1.05 or
                  # err > self._err_before*self.dwidth1 or
                  err > self._err_before*1.05 or
                  self._err_inc_count > 2):

                self.set_damping(self.damping*self.dwidth2)
                self._err_guidance = self._err_guidance*self.dwidth1
                # this makes reduction of damping milder...not sure if this is better....
                # self._err_guidance = self._err_before

                self.copyback_x(X[0], self._solbackup)
                self.update_x(self._delta)
                self._kiter = self._kiter - 1
                self._err_inc_count = 0

                if self.damping < self.minimum_damping:
                    # Let's give up ...(sad face)
                    self._done = True
                else:
                    dprint1("new damping (reduced), ref_error, current_error",
                            self.damping, self._err_before, err)

                    self._err_before = err
                    if self.scheme_name != "fixed-point":
                        return

            elif err < self._err_guidance*self.dwidth2 and self.damping < 1.0:
                self._err_guidance = err
                self._err_before = err
                self.set_damping(self.damping*self.dwidth1)
#                self.set_damping(self.damping*1.2)
                dprint1("new damping (increased)", self.damping)

            else:
                if err > self._err_before:
                    self._err_inc_count = self._err_inc_count + 1
                else:
                    self._err_inc_count = 0
                self._err_before = err
                self._err_guidance = self._err_guidance*1.02

            self.error_record.append(err)

            if self._kiter >= self._maxiter:
                self._done = True

        stopit = self.call_dwc_nliteration()
        if stopit:
            self._done = True

        self.residual_record.append(residual)

        if self.kiter > 5 and residual < 3e-6:
            self._done = True

        if not self._converged and not self._done:
            self.damping_record.append(self.damping)

            self._solbackup = self.copy_x(X[0])
            self._delta = self.do_solve(update_operator=update_operator)
            self.update_x(self._delta)
            self.engine.add_FESvariable_to_NS(self.get_phys())

            if self._kiter >= self._maxiter:
                self._done = True

        if self._done and not stopit:
            if self.damping != 1.0 and self.damping > self.minimum_damping:
                self.set_damping(1.0)
                self._done = False
                self._converged = False
                self._fixed_damping = True

        if self._done:
            if self._converged:
                dprint1("converged ("+self.scheme_name + ") #iter=", self.kiter)
                dprint1("final error |err| = ", err)
                dprint1("damping parameters", self.damping_record)

            else:
                dprint1("no convergence ("+self.scheme_name+" interation)")
                dprint1("damping parameters", self.damping_record)

            if self.verbose:
                dprint1("err history = ", self.error_record)
                dprint1("err^2 history (decomposition) = ", self.debug_data)
                dprint1("residuals", self.residual_record)

    def save_probe(self):
        from petram.mfem_config import use_parallel
        if use_parallel:
            from mpi4py import MPI
        else:
            from petram.helper.dummy_mpi import MPI
        myid = MPI.COMM_WORLD.rank

        if myid != 0:
            return

        from petram.sol.probe import Probe

        p1 = Probe(self.gui.name()+'_error')
        p2 = Probe(self.gui.name()+'_damping')

        for i, v in enumerate(self.error_record):
            p1.append_value(v, t=i)
        for i, v in enumerate(self.damping_record):
            p2.append_value(v, t=i)

        p1.write_file()
        p2.write_file()

        NonlinearBaseSolver.save_probe(self)

    def adjust_operator(self, A, M, mask):
        '''
        Add Jacobian term to operator
        '''
        from petram.engine import max_matrix_num

        M = (M[max_matrix_num//2]).get_subblock(mask[0], mask[1])

        if self.kiter == 0:
            pass
        else:
            A = A + M
        return A


class FixedPointSolver(NewtonSolver):
    def __init__(self, *args, **kwargs):
        NewtonSolver.__init__(self, *args, **kwargs)
        self.scheme_name = "fixed-point"

    def assemble(self, inplace=True, update=False):
        NonlinearBaseSolver.assemble(self, inplace=inplace, update=update)

    def adjust_operator(self, A, M, mask):
        return A



'''

   Weakform : interface to use MFEM integrator

'''
from petram.phys.coefficient import SCoeff, VCoeff, DCoeff, MCoeff
from petram.phys.vtable import VtableElement, Vtable
from petram.mfem_config import use_parallel
import os
import numpy as np
from abc import abstractmethod

from petram.phys.phys_model import Phys
from petram.model import Domain, Bdry, Edge, Point, Pair

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('WeakForm')

if use_parallel:
    import mfem.par as mfem
else:
    import mfem.ser as mfem


def get_integrators(filename, return_all=False):
    import petram.engine
    fid = open(os.path.join(os.path.dirname(
        petram.engine.__file__), 'data', filename), 'r')
    lines = fid.readlines()
    fid.close()

    lines = [l for l in lines if not l.startswith('#')]
    lines = [l.split("|")[1:] for l in lines]
    names = [l[0].strip() for l in lines]
    domains = [[x.strip() for x in l[1].split(',') if x.strip() != '']
               for l in lines]
    ranges = [[x.strip() for x in l[2].split(',') if x.strip() != '']
              for l in lines]
    for trial, test in zip(domains, ranges):
        if len(test) == 0:
            for x in trial:
                test.append(x)

    def x(txt):
        xx = [x.strip() for x in txt.split(",")]
        return xx

    coeffs = [x(l[3]) for l in lines]
    dims = [[int(x.strip()[0]) for x in l[6].split(',')] for l in lines]
    wf_forms = [l[4] for l in lines]
    strong_forms = [l[5] for l in lines]

    if return_all:
        return list(zip(names, domains, ranges, coeffs, dims, wf_forms, strong_forms))
    else:
        return list(zip(names, domains, ranges, coeffs, dims))


bilinintegs = get_integrators('BilinearOps')
linintegs = get_integrators('LinearOps')

data = [("coeff_lambda", VtableElement("coeff_lambda", type='array',
                                       guilabel="lambda", default=0.0, tip="coefficient",))]


class WeakIntegration(Phys):
    vt_coeff = Vtable(data)

    def attribute_set(self, v):
        v['use_src_proj'] = False
        v['use_dst_proj'] = False
        v['coeff_type'] = 'Scalar'
        v['integrator'] = 'MassIntegrator'
        v['test_idx'] = 0  # (index)
        self.vt_coeff.attribute_set(v)
        v = Phys.attribute_set(self, v)
        return v

    def get_panel1_value(self):
        dep_vars = self.get_root_phys().dep_vars
        return [dep_vars[self.test_idx],
                self.coeff_type,
                self.vt_coeff.get_panel_value(self)[0],
                self.integrator, ]
        # self.use_src_proj,
        # self.use_dst_proj,]

    def panel1_tip(self):
        pass

    def itg_choice_cb(self):
        names = [x[0] for x in self.itg_choice()]
        return names

    def panel1_param(self):
        import wx
        p = ["coeff. type", "Scalar", 4,
             {"style": wx.CB_READONLY, "choices": ["Scalar", "Vector", "Diagonal", "Matrix"]}]

        names = [x[0] for x in self.itg_choice()]

        import wx
        from petram.pi.widget_forms import WidgetForms

        p2 = ["integrator", None, 99,
              {"UI": WidgetForms,
               # "span": (1, 2),
               # "choices": names,
               "choices_cb": self.itg_choice_cb, }]

        dep_vars = self.get_root_phys().dep_vars
        panels = self.vt_coeff.panel_param(self)
        ll = [["test space (Rows)", dep_vars[0], 4,
               {"style": wx.CB_READONLY, "choices": dep_vars}],
              p,
              panels[0],
              p2, ]
        #["use src proj.",  self.use_src_proj,   3, {"text":""}],
        # ["use dst proj.",  self.use_dst_proj,   3, {"text":""}],  ]
        return ll

    def is_complex(self):
        return self.get_root_phys().is_complex_valued

    def import_panel1_value(self, v):
        dep_vars = self.get_root_phys().dep_vars
        self.test_idx = dep_vars.index(str(v[0]))
        self.coeff_type = str(v[1])
        self.vt_coeff.import_panel_value(self, (v[2],))
        self.integrator = str(v[3])
        #self.use_src_proj = v[4]
        #self.use_dst_proj = v[5]

    def preprocess_params(self, engine):
        self.vt_coeff.preprocess_params(self)
        super(WeakIntegration, self).preprocess_params(engine)

    def add_contribution(self, engine, a, real=True, is_trans=False, is_conj=False):
        c = self.vt_coeff.make_value_or_expression(self)[0]

        if real:
            dprint1("Add "+self.integrator + " contribution(real)" +
                    str(self._sel_index), "c", c)
        else:
            dprint1("Add "+self.integrator + " contribution(imag)" +
                    str(self._sel_index), "c", c)

        cotype = self.coeff_type[0]
        use_dual = False
        for b in self.itg_choice():
            if b[0] == self.integrator:
                use_dual = "S*2" in b[3]
                break

        c_coeff = self.get_coefficient_from_expression(c, cotype,
                                                       use_dual=use_dual,
                                                       real=real,
                                                       is_conj=is_conj)

        if self.integrator == 'DerivativeIntegrator1':
            integrator = getattr(mfem, 'DerivativeIntegrator')
            c_coeff = (c_coeff, 0)
        elif self.integrator == 'DerivativeIntegrator2':
            integrator = getattr(mfem, 'DerivativeIntegrator')
            c_coeff = (c_coeff, 1)
        elif self.integrator == 'DerivativeIntegrator3':
            integrator = getattr(mfem, 'DerivativeIntegrator')
            c_coeff = (c_coeff, 2)
        else:
            integrator = getattr(mfem, self.integrator)

        if isinstance(self, Bdry):
            # print "Bdry Integrator"
            adder = a.AddBoundaryIntegrator
        elif isinstance(self, Domain):
            # print "Domain Integrator"
            adder = a.AddDomainIntegrator
        else:
            assert False, "this class is not supported in weakform"
        self.add_integrator(engine, 'c', c_coeff,
                            adder, integrator, transpose=is_trans)

    def add_bf_contribution(self, engine, a, real=True, kfes=0):
        self.add_contribution(engine, a, real=real)

    def add_lf_contribution(self, engine, b, real=True, kfes=0):
        self.add_contribution(engine, b, real=real)

    def add_mix_contribution2(self, engine, mbf, r, c,  is_trans, is_conj,
                              real=True):
        self.add_contribution(engine, mbf, real=real,
                              is_trans=is_trans, is_conj=is_conj)

    @abstractmethod
    def itg_choice(self):
        return []


class WeakLinIntegration(WeakIntegration):
    def has_lf_contribution(self, kfes):
        return self.test_idx == kfes

    def itg_choice(self):
        t = self.get_root_phys().get_fec_type(self.test_idx)
        if len(t) > 2 and t[2] == "v":
            t = t[:3]
        bb = [b for b in linintegs if t in b[1]]

        if (t == 'RT' and self.get_root_phys().dim == 2 and
                self.get_root_phys().geom_dim == 3):
            t = 'L2'
            bb = bb + [b for b in linintegs if t in b[1]]

        dim = self.dim
        bb = [b for b in bb if dim in b[4]]

        return bb


class WeakBilinIntegration(WeakIntegration):
    def itg_choice(self):
        t = self.get_root_phys().get_fec_type(self.test_idx)
        if len(t) > 2 and t[2] == "v":
            t = t[:3]
        bb = [b for b in bilinintegs if t in b[2]]

        # adjust choices for RT-Trace (I need this to have MassIntegrator in menu)
        if (t == 'RT' and self.get_root_phys().dim == 2 and
                self.get_root_phys().geom_dim == 3):
            t = 'L2'
            bb = bb + [b for b in bilinintegs if t in b[2]]

        if self.paired_var is not None:
            paired_name, paired_idx = self.paired_var
        else:
            paired_name = self.get_root_phys().name()
            paired_idx = 0

        if paired_name in self.get_root_phys().parent:
            phys1 = (self.get_root_phys().parent)[paired_name]
            fes_idx1 = paired_idx

            from petram.helper.projection import find_fes_mapping
            t2, _order = find_fes_mapping(phys1, fes_idx1,
                                          self.get_root_phys(), self.test_idx)
            #t2 = (self.get_root_phys().parent)[paired_name].get_fec_type(paired_idx)
            if len(t2) > 2 and t2[2] == "v":
                t2 = t2[:3]
            else:
                t2 = t2[:2]
            bb2 = [b for b in bb if t2 in b[1]]

            # adjust choices for RT-Trace
            if (t2 == 'RT' and self.get_root_phys().dim == 2 and
                    self.get_root_phys().geom_dim == 3):
                t2 = 'L2'
                bb2 = bb2 + [b for b in bb if t2 in b[1]]
        else:
            bb2 = bb
        dim = self.dim

        bb = [b for b in bb2 if dim in b[4]]

        return bb

    def attribute_set(self, v):
        v = super(WeakBilinIntegration, self).attribute_set(v)
        v['paired_var'] = None  # (phys_name, index)
        v['use_symmetric'] = False
        v['use_conj'] = False
        return v

    def get_panel1_value(self):
        '''
        (old routine will be deleted)
        if self.paired_var is not None:
            try:
                mfem_physroot = self.get_root_phys().parent
                var_s = mfem_physroot[self.paired_var[0]].dep_vars
                n = var_s[self.paired_var[1]]
                p = self.paired_var[0]
            except:
                self.paired_var = None

        if self.paired_var is None:
            n = self.get_root_phys().dep_vars[0]
            p = self.get_root_phys().name()

        var = n + " ("+p + ")"
        '''
        from petram.utils import pv_get_gui_value
        gui_value, self.paired_var = pv_get_gui_value(self, self.paired_var)

        v1 = [gui_value]
        v2 = super(WeakBilinIntegration, self).get_panel1_value()
        v3 = [self.use_symmetric, self.use_conj]
        return v1 + v2 + v3

    def panel1_tip(self):
        pass

    def is_complex(self):
        return self.get_root_phys().is_complex_valued

    def import_panel1_value(self, v):
        '''
        (old routine will be deleted)
        mfem_physroot = self.get_root_phys().parent
        names, pnames, pindex = mfem_physroot.dependent_values()

        if len(v[0]) == 0:
            # v[0] could be '' if object is based to a tree.
            idx = 0
        else:
            idx = names.index(str(v[0]).split("(")[0].strip())

        self.paired_var = (pnames[idx], pindex[idx])
        '''
        from petram.utils import pv_from_gui_value
        self.paired_var = pv_from_gui_value(self, v[0])

        super(WeakBilinIntegration, self).import_panel1_value(v[1:-2])
        self.use_symmetric = v[-2]
        self.use_conj = v[-1]

    def panel1_param(self):
        '''
        (old routine will be deleted)
        import wx
        mfem_physroot = self.get_root_phys().parent
        names, pnames, pindex = mfem_physroot.dependent_values()
        names = [n+" ("+p + ")" for n, p in zip(names, pnames)]

        ll1 = [["paired variable", "S", 4,
                {"style": wx.CB_READONLY, "choices": names}]]
        '''
        from petram.utils import pv_panel_param
        ll1 = [pv_panel_param(self, "paird var."), ]

        ll2 = super(WeakBilinIntegration, self).panel1_param()
        ll3 = [["make symmetric",  self.use_symmetric,   3, {"text": ""}],
               ["use  conjugate",  self.use_conj,   3, {"text": ""}], ]
        ll = ll1 + ll2 + ll3

        return ll

    def has_bf_contribution(self, kfes):
        mfem_physroot = self.get_root_phys().parent
        var_s = mfem_physroot[self.paired_var[0]].dep_vars
        trialname = var_s[self.paired_var[1]]
        testname = self.get_root_phys().dep_vars[self.test_idx]

        return (trialname == testname) and (self.test_idx == kfes)

    def has_mixed_contribution(self):
        mfem_physroot = self.get_root_phys().parent
        var_s = mfem_physroot[self.paired_var[0]].dep_vars
        trialname = var_s[self.paired_var[1]]
        testname = self.get_root_phys().dep_vars[self.test_idx]

        return (trialname != testname)

    def get_mixedbf_loc(self):
        mfem_physroot = self.get_root_phys().parent
        var_s = mfem_physroot[self.paired_var[0]].dep_vars
        trialname = var_s[self.paired_var[1]]
        testname = self.get_root_phys().dep_vars[self.test_idx]

        loc = []
        is_trans = 1
        loc.append((testname, trialname, is_trans, 1))

        if self.use_symmetric and not self.use_conj:
            loc.append((testname, trialname, -1, -1))
        if self.use_symmetric and self.use_conj:
            loc.append((testname, trialname, -1, 1))

        return loc

    def get_projection(self):
        return 1


def add_delta_contribution(obj, engine, a, real=True, is_trans=False, is_conj=False):
    self = obj
    c = self.vt_coeff.make_value_or_expression(self)[0]
    if isinstance(c, str):
        c = [c]

    if real:
        dprint1("Add "+self.integrator + " delta (real)" +
                str(self._sel_index), "c", c)
    else:
        dprint1("Add "+self.integrator + " delta (imag)" +
                str(self._sel_index), "c", c)

    cotype = self.coeff_type[0]

    if self.get_root_phys().vdim > 1:
        dim = self.get_root_phys().vdim
    else:
        el_name = self.get_root_phys().element
        dim = self.get_root_phys().geom_dim
    sdim = self.get_root_phys().geom_dim

    integrator = getattr(mfem, self.integrator)
    adder = a.AddDomainIntegrator

    for pos in self.pos_value:
        args = list(pos[:sdim])
        if cotype == 'S':
            for b in self.itg_choice():
                if b[0] == self.integrator:
                    break

            if not "S*2" in b[3]:
                if isinstance(c[0], str):
                    c_coeff = None
                    # c_coeff = SCoeff(c[0],  self.get_root_phys().ind_vars,
                    #                 self._local_ns, self._global_ns,
                    #                 real = real, conj=is_conj)
                    value = eval(c[0], self._global_ns, self._local_ns)
                    dprint1("time dependent delta", value)
                    args.append(float(value))
                else:
                    c_coeff = None
                    if real:
                        args.append(float(np.array(c)[0].real))
                    else:
                        args.append(float(np.array(c)[0].imag))
                if args[-1] != 0:
                    d = mfem.DeltaCoefficient(*args)
                    if c_coeff is not None:
                        assert False, "This option needs update of PyMFEM"
                        d2 = mfem.ProductCoefficient(c_coeff, d)
                        d2._linked_c = (c_coeff, d)
                        adder(integrator(d2))
                    else:
                        adder(integrator(d))

            else:  # so far this is only for an elastic integrator
                if real:
                    args.append(float(np.array(c)[0].real))
                else:
                    args.append(float(np.array(c)[0].imag))
                d1 = mfem.DeltaCoefficient(*args)
                if real:
                    args.append(float(np.array(c)[1].real))
                else:
                    args.append(float(np.array(c)[1].imag))
                d2 = mfem.DeltaCoefficient(*args)
                adder(integrator(d1, d2))

        elif cotype == 'V':
            if real:
                direction = np.array(c).real
            else:
                direction = np.array(c).imag
            args.append(1.0)
            dir = mfem.Vector(direction)
            d = mfem.VectorDeltaCoefficient(dir, *args)
            adder(integrator(d))
        else:
            assert False, "M and D are not supported for delta coefficient"


def validate_sel(value, obj, w):
    g = obj._global_ns
    try:
        value = eval(value, g)
        pos = np.atleast_2d(value)
        return True
    except:
        return False


class WeakLinDeltaIntegration(WeakLinIntegration):
    def panel2_param(self):
        return [["Position",  "[[0,0,0],]",  0, {'changing_event': True,
                                                 'setfocus_event': True,
                                                 'validator': validate_pos,
                                                 'validator_param': self}]]

    def import_panel2_value(self, v):
        self.sel_index = ['all']
        self.sel_index_txt = '["all"]'
        self.pos_txt = v[0]
        g = self._global_ns
        try:
            value = eval(v[0], g)
            self.pos_value = np.atleast_2d(value)
        except:
            import traceback
            traceback.print_exc()
            pass

    def attribute_set(self, v):
        v = super(WeakLinDeltaIntegration, self).attribute_set(v)
        v['pos_txt'] = "[[ 0,0,0,],]"
        v['pos_value'] = np.array([[0, 0, 0, ], ])
        return v

    def get_panel2_value(self):
        return (self.pos_txt,)

    def add_contribution(self, engine, a, real=True, is_trans=False, is_conj=False):
        g = self._global_ns
        try:
            value = eval(self.pos_txt, g)
            self.pos_value = np.atleast_2d(value)
        except:
            import traceback
            traceback.print_exc()
            raise
        add_delta_contribution(self, engine, a, real=real, is_trans=is_trans,
                               is_conj=is_conj)


class WeakBilinDeltaIntegration(WeakBilinIntegration):
    def panel2_param(self):
        return [["Position",  "[[0,0,0],]",  0, {'changing_event': True,
                                                 'setfocus_event': True,
                                                 'validator': validate_pos,
                                                 'validator_param': self}]]

    def import_panel2_value(self, v):
        self.sel_index = ['all']
        self.sel_index_txt = '["all"]'
        self.pos_txt = v[0]
        g = self._global_ns
        try:
            value = eval(v[0], g)
            self.pos_value = np.atleast_2d(value)
        except:
            import traceback
            traceback.print_exc()

    def attribute_set(self, v):
        v = super(WeakBilinDeltaIntegration, self).attribute_set(v)
        v['pos_txt'] = "[[ 0,0,0,],]"
        v['pos_value'] = np.array([[0, 0, 0, ], ])
        return v

    def get_panel2_value(self):
        return (self.pos_txt,)

    def add_contribution(self, engine, a, real=True, is_trans=False, is_conj=False):
        g = self._global_ns
        try:
            value = eval(self.pos_txt, g)
            self.pos_value = np.atleast_2d(value)
        except:
            import traceback
            traceback.print_exc()
            raise
        add_delta_contribution(self, engine, a, real=real, is_trans=is_trans,
                               is_conj=is_conj)



from petram.solver.std_solver_model import StdSolver, StandardSolver
import os
import numpy as np

from petram.model import Model
from petram.solver.solver_model import Solver
import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('StdMeshAdaptSolver')
rprint = debug.regular_print('StdMeshAdaptSolver')


class StdMeshAdaptSolver(StdSolver):
    @classmethod
    def fancy_menu_name(self):
        return 'AMR Stationary'

    @classmethod
    def fancy_tree_name(self):
        return 'AMR Stationary'

    def panel1_param(self):
        return [  # ["Initial value setting",   self.init_setting,  0, {},],
            ["physics model",   self.phys_model,  0, {}, ],
            ["initialize solution only", self.init_only,  3, {"text": ""}],
            [None,
             self.clear_wdir,  3, {"text": "clear working directory"}],
            [None,
             self.assemble_real,  3, {"text": "convert to real matrix (complex prob.)"}],
            [None,
             self.save_parmesh,  3, {"text": "save parallel mesh"}],
            [None,
             self.use_profiler,  3, {"text": "use profiler"}],
            ["indicator",   self.mesh_adapt_indicator,  0, {}],
            ["#mesh adapt",   self.mesh_adapt_num,  0, {}, ], ]

    def attribute_set(self, v):
        super(StdMeshAdaptSolver, self).attribute_set(v)
        v["mesh_adapt_indicator"] = ""
        v["mesh_adapt_num"] = 0
        return v

    def get_panel1_value(self):
        return (  # self.init_setting,
            self.phys_model,
            self.init_only,
            self.clear_wdir,
            self.assemble_real,
            self.save_parmesh,
            self.use_profiler,
            self.mesh_adapt_indicator,
            self.mesh_adapt_num)

    def import_panel1_value(self, v):
        #self.init_setting = str(v[0])
        self.phys_model = str(v[0])
        self.init_only = v[1]
        self.clear_wdir = v[2]
        self.assemble_real = v[3]
        self.save_parmesh = v[4]
        self.use_profiler = v[5]
        self.mesh_adapt_indicator = v[6]
        self.mesh_adapt_num = int(v[7])

    @debug.use_profiler
    def run(self, engine, is_first=True, return_instance=False):
        dprint1("Entering run", is_first, self.fullpath())
        if self.clear_wdir:
            engine.remove_solfiles()

        instance = StandardMeshAdaptSolver(self, engine)
        instance.set_blk_mask()
        if return_instance:
            return instance
        # We dont use probe..(no need...)
        # instance.configure_probes(self.probe)

        if self.init_only:
            engine.sol = engine.assembled_blocks[1][0]
            instance.sol = engine.sol
        else:
            if is_first:
                instance.assemble()
                is_first = False
            instance.solve()

        instance.save_solution(ksol=0,
                               skip_mesh=False,
                               mesh_only=False,
                               save_parmesh=self.save_parmesh)
        engine.sol = instance.sol
        dprint1(debug.format_memory_usage())
        return is_first


class StandardMeshAdaptSolver(StandardSolver):
    pass



'''
   CDict is a dictionary allowing to access items using class-like 
   syntax
'''
class tmp(object):
    pass

err_msg = 'cdict member name should be string, not starts from "_"'

class CDict(dict):
    def __getattr__(self, attr):
      # only attributes not starting with "_" are organinzed
      # in the tree
       if hasattr(self, attr):
           return dict.__getattr__(self, attr)
       if not attr.startswith("_"):
           try:
               return self.__getitem__(attr)
           except:
               raise
       raise AttributeError(err_msg)

    def __setattr__(self, attr, val):
       if hasattr(self, attr):
           return dict.__setattr__(self, attr, val)           
       if isinstance(attr, str):
           if not attr.startswith("_"):
               return self.__setitem__(attr, val)
       raise AttributeError(err_msg)           

    def __setitem__(self, key, value):
        ## check if key can be accessed as a class attribute
        try:
            if key.startswith('_'):
                 raise AttributeError(err_msg)
        except:
            raise AttributeError(err_msg)            
        o = tmp()
        try:
            exec("o."+key+"=3")
        except:
            raise AttributeError(err_msg)                                   
        dict.__setitem__(self, key, value)
    




import wx
from ifigure.utils.edit_list import TextCtrlCopyPaste

class TextCtrlCallBack(TextCtrlCopyPaste):
    def __init__(self, *args, **kwargs):
        self._previous_txt = ''
        setting = kwargs.pop('setting', {})
        self._callback = setting.pop('callback_method', None)
        callback_on_enter = setting.pop('callback_on_enter', False)
        args = list(args)
        if len(args) == 1: args.append(wx.ID_ANY)
        if len(args) == 2: args.append('')
        #kwargs['changing_event'] = True
        kwargs['style'] = wx.TE_PROCESS_ENTER
        TextCtrlCopyPaste.__init__(self, *args, **kwargs)

        if callback_on_enter:
            self.Bind(wx.EVT_TEXT_ENTER, self.onEnter)            
        
    def onKillFocus(self, evt):
        TextCtrlCopyPaste.onKillFocus(self, evt)
        if self._callback is not None:
            self._callback(evt)

    def onKeyPressed(self, evt):
        super(TextCtrlCallBack, self).onKeyPressed(evt)
        
    def onEnter(self, evt):
        super(TextCtrlCallBack, self).onEnter(evt)

    '''        
    def onKeyPressed(self, evt):
        print ("onKeyPress")
        TextCtrlCopyPaste.onKeyPressed(self, evt)
        if self._callback is not None:
            self._callback(evt)
    '''



from __future__ import print_function

import os
import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('Namespace')


class NSRef_mixin(object):
    hide_ns_menu = False

    def __init__(self, *args, **kwargs):
        object.__init__(self)
        self.reset_ns()

    def get_info_str(self):
        if self.ns_name is not None:
            return 'NS:'+self.ns_name
        return ""

    def reset_ns(self):
        if not hasattr(self, 'ns_name'):
            self.ns_name = None

    def get_ns_name(self):
        if not hasattr(self, 'ns_name'):
            self.reset_ns()
        return self.ns_name

    def find_ns_by_name(self):
        '''
        return NameSpace
        '''
        name = self.ns_name
        root = self.root()
        for obj in root.walk():
            if not isinstance(obj, NS_mixin):
                continue
            if obj.get_ns_name() == name and name is not None:
                return obj._global_ns
        return self.root()['General']._global_ns

    def find_nsobj_by_name(self):
        '''
        return model holding NameSpace for given name
        '''
        name = self.ns_name
        root = self.root()
        for obj in root.walk():
            if not isinstance(obj, NS_mixin):
                continue
            if obj.get_ns_name() == name and name is not None:
                return obj
        return self.root()['General']

    def new_ns(self, name):
        self.ns_name = name

    def delete_ns(self):
        self.ns_name = None


class NS_mixin(object):
    hide_ns_menu = False

    def __init__(self, *args, **kwargs):
        object.__init__(self)
        self.reset_ns()

    @property
    def namespace(self):
        return (self._global_ns, self._local_ns)

    def attribute_expr(self):
        '''
        define attributes evaluated as exprssion
        returns name and validator (float, int, complex or None)
        '''
        return [], []

    def attribute_mirror_ns(self):
        '''
        a list of attribute copied to ns
        '''
        return []

    def get_info_str(self):
        if self.ns_name is not None:
            return 'NS:'+self.ns_name
        return ""

    def reset_ns(self):
        self._global_ns = None
        self._local_ns = None
        self.ns_name = None
        self.ns_string = None
        self.dataset = None

    def get_ns_name(self):
        if not hasattr(self, 'ns_name'):
            self.reset_ns()
        return self.ns_name

    def find_ns_by_name(self, name):
        '''
        return NameSpace for given name
        It walks through the model tree to find an model object
        whose namespace name matches with input
        '''
        root = self.root()
        for obj in root.walk():
            if not isinstance(obj, NS_mixin):
                continue
            if obj.get_ns_name() == name:
                return obj, obj._global_ns
        return None, {}

    def get_ns_chain(self):
        chain = []
        p = self
        while p is not None:
            if isinstance(p, NS_mixin):
                if p.ns_name is not None:
                    chain.append(p)
#                elif len(p.get_default_ns()) != 0:
#                    chain.append(p)
            p = p.parent

        tmp = [x for x in reversed(chain)]
        gn = self.root()['General']
        if gn.get_ns_name() is not None:
            if len(tmp) == 0:
                tmp.append(gn)
            elif tmp[0] is not gn:
                tmp = [gn] + tmp
            else:
                pass
        return tmp

    def write_ns_script_data(self, dir=None):
        path1 = os.path.join(dir, self.ns_name+'_ns.py')
        path2 = os.path.join(dir, self.ns_name+'_ns.dat')
        fid = open(path1, 'w')
        fid.write(self.ns_string)
        fid.close()
        import petram.helper.pickle_wrapper as pickle
        fid = open(path2, 'wb')
        pickle.dump(self.dataset, fid)
        fid.close()

    def read_ns_script_data(self, dir=None):
        path1 = os.path.join(dir, self.ns_name+'_ns.py')
        path2 = os.path.join(dir, self.ns_name+'_ns.dat')
        fid = open(path1, 'r')
        #self.ns_string = '\n'.join(fid.readlines())
        self.ns_string = ''.join(fid.readlines())
        fid.close()
        import petram.helper.pickle_wrapper as pickle
        fid = open(path2, 'rb')
        self.dataset = pickle.load(fid)
        fid.close()

    def delete_ns(self):
        self._global_ns = None
        self._local_ns = None
        self.ns_name = None
        self.ns_string = None
        self.dataset = None

    def new_ns(self, name):
        self._global_ns = None
        self._local_ns = None
        self.ns_name = name
        self.ns_string = None
        self.dataset = None

    def preprocess_ns(self, ns_folder, data_folder):
        if self.get_ns_name() is None:
            return

        ns_script = ns_folder.get_child(name=self.ns_name+'_ns')
        if ns_script is None:
            raise ValueError("namespace script is not found")
        err_string = ns_script.reload_script()
        if err_string != '' and err_string is not None:
            assert False, err_string

        self.ns_string = ns_script._script._script

        data = data_folder.get_child(name=self.ns_name+'_data')
        if data is None:
            raise ValueError("dataset is not found")
        d = data.getvar()
        self.dataset = {k: d[k] for k in d}  # copy dict

    def get_default_ns(self):
        '''
        this method is overwriten when model wants to
        set its own default namespace. For example, when
        RF module set freq and omega
        '''
        return {}

    def eval_attribute_expr(self, targets=None):

        names, types = self.attribute_expr()
        exprs = [(x, x+'_txt', v) for x, v in zip(names, types)]

        invalid_expr = []
        result = {}
        for name, tname,  validator in exprs:
            if targets is not None and not tname in targets:
                continue
            try:
                void = {}
                x = eval(str(getattr(self, tname)), self._global_ns, void)
            except:
                import traceback
                traceback.print_exc()
                invalid_expr.append(tname)
                continue
            try:
                if validator is not None:
                    x = validator(x)
            except:
                import traceback
                traceback.print_exc()
                invalid_expr.append(tname)
                continue
            result[name] = x

        return result, invalid_expr

    def eval_ns(self):
        chain = self.get_ns_chain()

        if not self.is_enabled():
            # if it is not enabled we use default _global_ns
            if chain[0] is not self:
                self._global_ns = chain[0]._global_ns
                self._local_ns = self.root()._variables
                return
            else:
                assert False, "General should not be disabled"

        l = self.get_default_ns()

        from petram.helper.variables import var_g
        g = var_g.copy()

        import mfem
        if mfem.mfem_mode == 'serial':
            g['mfem'] = mfem.ser
        elif mfem.mfem_mode == 'parallel':
            g['mfem'] = mfem.par
        else:
            assert False, "PyMFEM is not loaded"

        import numpy
        g['np'] = numpy
        from petram.helper.variables import variable, coefficient
        g['variable'] = variable
        g['coefficient'] = coefficient

        if self.root() is self:
            if not hasattr(self.root(), "_variables"):
                from petram.helper.variables import Variables
                self.root()._variables = Variables()
        else:
            self._local_ns = self.root()._variables

        if len(chain) == 0:
            raise ValueError("namespace chain is not found")
        # step1 (fill ns using upstream + constant (no expression)
        if chain[-1] is not self:
           # if len(l) == 0:
            self._global_ns = chain[-1]._global_ns
            for k in l:
                self._global_ns[k] = l[k]
            g = self._global_ns
            #self._local_ns = chain[-1]._local_ns
           # else:
           #     self._global_ns = g
           #     for k in l:
           #         g[k] = l[k]
           #     for k in chain[-1]._global_ns:
           #         g[k] = chain[-1]._global_ns[k]
            #self._local_ns = {}
        elif len(chain) > 1:
            # step 1-1 evaluate NS chain except for self and store dataset to
            # g including mine
            self._global_ns = g
            for p in chain[:-1]:  # self.parents:
                if not isinstance(p, NS_mixin):
                    continue
                ll = p.get_default_ns()
                if (p.ns_string == '' or p.ns_string is None and
                        len(ll) == 0):
                    continue
                for k in ll:
                    g[k] = ll[k]
                if p.ns_name is not None:
                    try:
                        if p.dataset is not None:
                            for k in p.dataset:
                                g[k] = p.dataset[k]
                        for k in p.attribute_mirror_ns():
                            g[k] = chain[-2]._global_ns[k]
                        if (p.ns_string != '' and p.ns_string is not None):
                            #exec(p.ns_string, g, ll)
                            #for k in ll: g[k] = ll[k]
                            exec(p.ns_string, g)

                    except Exception as e:
                        import traceback
                        assert False, traceback.format_exc()
            if self.dataset is not None:
                for k in self.dataset:
                    g[k] = self.dataset[k]
        else:
            self._global_ns = g
            for k in l:
                g[k] = l[k]
            if self.dataset is not None:
                for k in self.dataset:
                    g[k] = self.dataset[k]
        # step2 eval attribute using upstream + non-expression
        result, invalid = self.eval_attribute_expr()
        for k in result:
            setattr(self, k, result[k])

        # step 3 copy attributes to ns
        attrs = self.attribute_mirror_ns()
        for a in attrs:
            if not a in invalid:
                g[a] = getattr(self, a)

        # step 4 run namespace scripts otherise exit
        for k in l:
            g[k] = l[k]  # copying default ns

        try:
            l = {}
            if (self.ns_string != '' and self.ns_string is not None):
                #exec(self.ns_string, g, l)
                exec(self.ns_string, g)
            else:
                pass  # return
        except Exception as e:
            import traceback
            assert False, traceback.format_exc()

        # 2021.08.25. passing g only above allows for list comprehension to work.
        # for k in l:
        #     g[k] = l[k]

        # step 5  re-eval attribute with self-namespace
        #         passing previous invalid as a list of variables
        #         to evaluate
        result, invalid = self.eval_attribute_expr(invalid)
        for k in result:
            setattr(self, k, result[k])

        # if self is not self.root()["General"] (Let's set it in General too)
        from petram.helper.dot_dict import DotDict
        g['general'] = DotDict(self.root()["General"]._global_ns)

        # if something is still not known,,, raise
        if len(invalid) != 0:
            raise ValueError(
                "failed to evaluate variable " + ', '.join(invalid))

    # parameters with validator

    def check_param_expr(self, value, param, ctrl):
        try:
            self.eval_param_expr(str(value), param)
            return True
        except:
            import petram.debug
            import traceback
            if petram.debug.debug_default_level > 2:
                traceback.print_exc()
            return False

    def eval_param_expr(self, value, param):
        x = eval(value, self._global_ns, self._local_ns)
        dprint2('Value Evaluation ', param, '=', x)
        return x, None

    # note that physics modules overwrite this with more capablie version
    def make_param_panel(self, base_name, value):
        return [base_name + "(=)",  value, 0,
                {'validator': self.check_param_expr,
                 'validator_param': base_name}]



from __future__ import print_function

import numpy as np
import scipy
import scipy.sparse.linalg

def null(a, rtol=1e-5):
    ''' 
    find null space of matrix a
    '''
    u, s, v = np.linalg.svd(a)
    rank = (s > rtol*s[0]).sum()
    return s, rank, v[rank:].T.copy()

# this one does not work...
def nulls(a, rtol=1e-5):
    d = min(a.shape)-1
    u, s, v = scipy.sparse.linalg.svds(a, k=d)
    print(v.shape)
    null_space = v.conj().ravel()
    rank = (s > rtol*s[0]).sum()    
    return s, rank, null_space

def nulls2(a, rtol=1e-12):
    from sparsesvd import sparsesvd
    smat = scipy.sparse.csc_matrix(a)
    ut, s, vt = sparsesvd(smat, np.min(a.shape))
    print(vt.shape)
    padding = max(0,max(np.shape(a))-np.shape(s)[0])
    null_mask = np.concatenate(((s <= rtol), np.ones((padding,),dtype=bool)),axis=0)
    print(null_mask.shape)
    null_space = scipy.compress(null_mask, vt, axis=0)
    rank = (s > rtol*s[0]).sum()
    return s, rank, scipy.transpose(null_space)

def add_constraints(A, b, mm, m=None):
    '''
     A   , [mm, 0]^t
 
    [mm 0]  [0]
    '''
    from scipy.sparse import csr_matrix, bmat, hstack
    asize = A.shape[1]
    mmsize = mm.shape[1]
    delta = asize - mmsize
    if delta > 0:
        mm = hstack((mm, csr_matrix((mm.shape[0], delta), dtype = A.dtype)), format = 'csr')
    d = csr_matrix((mm.shape[0], mm.shape[0]), dtype = A.dtype)
    if m is None:
        m = np.array([0]*mm.shape[0])

    return bmat([[A, mm.transpose()], [mm, d]], format = 'csr'), np.hstack((b, m))

def make_numpy_coo_matrix(A):
    '''
    solve A*x= b using mumps through PETSc
    '''
    print("!!!!! Deprecated use mfem.commons.sparse_utils.sparsemat_to_scipycsr")
    I = A.GetIArray()
    J = A.GetJArray()
    D = A.GetDataArray()
    return scipy.sparse.csr_matrix((D, J, I), (A.Size(),A.Size())).tocoo(False)

def gather_vector(fespace, A, data,  mpi_data_type, assemble=False ):
    '''
    gather vector to root node. 
    A: HypreParMatrix to know the total length of data
    B: Vector to be collected 
    '''
    from mpi4py import MPI
    myid     = MPI.COMM_WORLD.rank

    # Need to collect RHS to call mumps
    #disps = fespace.GetGlobalTDofNumber(0)
    #disps = MPI.COMM_WORLD.gather(disps, root = 0)
    #rcounts = fespace.GetTrueVSize()
    rcounts = data.shape[0]
    rcounts = MPI.COMM_WORLD.gather(rcounts, root = 0)
    cm = np.hstack((0, np.cumsum(rcounts)))
    disps = list(cm[:-1])        
    recvdata = None
    senddata = [data, data.shape[0]]
    
    if myid ==0:
       length =  cm[-1]
       recvbuf = np.empty([length], dtype=data.dtype)
       recvdata = [recvbuf, rcounts, disps, mpi_data_type]       
    else:
       recvdata = [None, rcounts, disps, mpi_data_type]
       recvbuf = None
    MPI.COMM_WORLD.Barrier()           
    MPI.COMM_WORLD.Gatherv(senddata, recvdata,  root = 0)
    if myid == 0:
        MPI.COMM_WORLD.Barrier()
        return np.array(recvbuf)
    MPI.COMM_WORLD.Barrier()
    return None
    

def scatter_vector(fespace, vector, mpi_data_type):
    from mpi4py import MPI    
    # Need to scatter solution
    senddata = None
    #disps = fespace.GetGlobalTDofNumber(0)
    #disps = MPI.COMM_WORLD.gather(disps, root = 0)
    rcounts = fespace.GetTrueVSize()
    rcounts = MPI.COMM_WORLD.gather(rcounts, root = 0)
    disps = list(np.hstack((0, np.cumsum(rcounts)))[:-1])
    recvdata = np.empty([fespace.GetTrueVSize()], dtype="float64")
    if vector is not None: 
        sol = np.array(vector, dtype="float64")
        senddata = [sol, rcounts, disps, mpi_data_type]
    MPI.COMM_WORLD.Scatterv(senddata, recvdata, root = 0)
    MPI.COMM_WORLD.Barrier()        
    return list(recvdata)    



'''

  evaulate derivative of Nedelec element using
  DiscreteLinearOperator

'''
from petram.mfem_config import use_parallel
import petram.debug
dprint1, dprint2, dprint3 = petram.debug.init_dprints('eval_deriv')

if use_parallel:
   import mfem.par as mfem
   FiniteElementSpace = mfem.ParFiniteElementSpace
   DiscreteLinearOperator = mfem.ParDiscreteLinearOperator
   GridFunction = mfem.ParGridFunction
   getFESpace = 'ParFESpace'
   getMesh = 'GetParMesh'   
else:
   import mfem.ser as mfem
   FiniteElementSpace = mfem.FiniteElementSpace
   DiscreteLinearOperator = mfem.DiscreteLinearOperator
   GridFunction = mfem.GridFunction
   getFESpace = 'FESpace'   
   getMesh = 'GetMesh'


def eval_curl(gfr, gfi = None):
    '''
    evaluate curl 
    '''
    fes = getattr(gfr, getFESpace)()
    ordering = fes.GetOrdering()
    mesh = getattr(fes, getMesh)()
    vdim = 1
    sdim = mesh.SpaceDimension()
    p = fes.GetOrder(0)
    rt_coll = mfem.RT_FECollection(p-1, sdim)

    rts = FiniteElementSpace(mesh,  rt_coll, vdim, ordering)
    
    curl = DiscreteLinearOperator(fes, rts)
    itp = mfem.CurlInterpolator()
    curl.AddDomainInterpolator(itp)
    curl.Assemble();
    curl.Finalize();

    br   = GridFunction(rts)    
    curl.Mult(gfr,  br)
    if gfi is not None:
       bi = GridFunction(rts)           
       curl.Mult(gfi,  bi)
    else:
       bi = None
    ### needs to return rts to prevent rts to be collected.
    return br, bi, (rt_coll, rts)
 
def eval_grad(gfr, gfi = None):
    '''
    evaluate gradient
    '''
    fes = getattr(gfr, getFESpace)()
    ordering = fes.GetOrdering()
    mesh = getattr(fes, getMesh)()
    vdim = 1
    sdim = mesh.SpaceDimension()
    p = fes.GetOrder(0)
    nd_coll = mfem.ND_FECollection(p, sdim)

    nds = FiniteElementSpace(mesh,  nd_coll, vdim, ordering)
    
    grad = DiscreteLinearOperator(fes, nds)
    itp = mfem.GradientInterpolator()
    grad.AddDomainInterpolator(itp)
    grad.Assemble();
    grad.Finalize();

    br   = GridFunction(rts)    
    grad.Mult(gfr,  br)
    if gfi is not None:
       bi = GridFunction(rts)           
       grad.Mult(gfi,  bi)
    else:
       bi = None
    ### needs to return rts to prevent rts to be collected.
    return br, bi, (rt_coll, nds)
 
def eval_div(gfr, gfi = None):
    '''
    evaluate divergence
    '''
    fes = getattr(gfr, getFESpace)()
    ordering = fes.GetOrdering()
    mesh = getattr(fes, getMesh)()
    vdim = 1
    sdim = mesh.SpaceDimension()
    p = fes.GetOrder(0)
    l2_coll = mfem.L2_FECollection(p, 1)

    l2s = FiniteElementSpace(mesh,  l2_coll, vdim, ordering)
    
    div = DiscreteLinearOperator(fes, l2s)
    itp = mfem.DivergenceInterpolator()
    div.AddDomainInterpolator(itp)
    div.Assemble();
    div.Finalize();

    br   = GridFunction(rts)    
    div.Mult(gfr,  br)
    if gfi is not None:
       bi = GridFunction(rts)           
       div.Mult(gfi,  bi)
    else:
       bi = None
    ### needs to return rts to prevent rts to be collected.
    return br, bi, (rt_coll, l2s)





'''
   IntegralEvaluator:
      a thing to evaluate integral on a boundary/domain
'''
import numpy as np
import weakref
import six

from weakref import WeakKeyDictionary as WKD
from weakref import WeakValueDictionary as WVD


from petram.mfem_config import use_parallel
if use_parallel:
    import mfem.par as mfem
    from mfem.par import GlobGeometryRefiner as GR    
else:
    import mfem.ser as mfem
    from mfem.ser import GlobGeometryRefiner as GR
    
from petram.sol.evaluator_agent import EvaluatorAgent
Geom = mfem.Geometry()

def do_integration(expr, solvars, phys, mesh, kind, attrs,
                   order, num):
    from petram.helper.variables import (Variable,
                                         var_g,
                                         NativeCoefficientGenBase,
                                         CoefficientVariable)
    from petram.phys.coefficient import SCoeff

    code = compile(expr, '<string>', 'eval')
    names = code.co_names

    g = {}
    #print solvars.keys()
    for key in phys._global_ns.keys():
       g[key] = phys._global_ns[key]
    for key in solvars.keys():
       g[key] = solvars[key]

    l = var_g.copy()

    ind_vars = ','.join(phys.get_independent_variables())

    if kind == 'Domain':
        size = max(max(mesh.attributes.ToList()), max(attrs))
    else:
        size = max(max(mesh.bdr_attributes.ToList()), max(attrs))
        
    arr = [0]*(size)
    for k in attrs:
        arr[k-1] = 1
    flag = mfem.intArray(arr)

    s = SCoeff(expr, ind_vars, l, g, return_complex=False)

    ## note L2 does not work for boundary....:D
    if kind == 'Domain':
        fec = mfem.L2_FECollection(order, mesh.Dimension())
    else:
        fec = mfem.H1_FECollection(order, mesh.Dimension())

    fes = mfem.FiniteElementSpace(mesh, fec)
    one = mfem.ConstantCoefficient(1)

    gf = mfem.GridFunction(fes)
    gf.Assign(0.0)

    if kind == 'Domain':
        gf.ProjectCoefficient(mfem.RestrictedCoefficient(s, flag))
    else:
        gf.ProjectBdrCoefficient(mfem.RestrictedCoefficient(s, flag), flag)

    b = mfem.LinearForm(fes)
    one = mfem.ConstantCoefficient(1)
    if kind == 'Domain':
        itg = mfem.DomainLFIntegrator(one)
        b.AddDomainIntegrator(itg)
    else:
        itg = mfem.BoundaryLFIntegrator(one)
        b.AddBoundaryIntegrator(itg)

    b.Assemble()

    from petram.engine import SerialEngine
    en = SerialEngine()
    ans = mfem.InnerProduct(en.x2X(gf), en.b2B(b))
    
    if not np.isfinite(ans):
        print("not finite", ans, arr)
        print(size, mesh.bdr_attributes.ToList())
        from mfem.common.chypre import LF2PyVec, PyVec2PyMat, Array2PyVec, IdentityPyMat
        #print(list(gf.GetDataArray()))
        print(len(gf.GetDataArray()), np.sum(gf.GetDataArray()))
        print(np.sum(list(b.GetDataArray())))
    return ans

class IntegralEvaluator(EvaluatorAgent):
    def __init__(self, battrs, decimate=1):
        super(IntegralEvaluator, self).__init__()
        self.battrs = battrs
        self.decimate = decimate

    def eval_integral(self, expr, solvars, phys,
                      kind='domain', attrs='all', order=2, num=-1):

        from .bdr_nodal_evaluator import get_emesh_idx

        emesh_idx = get_emesh_idx(self, expr, solvars, phys)

        if len(emesh_idx) > 1:
            assert False, "expression involves multiple mesh (emesh length != 1)"

        mesh = self.mesh()[emesh_idx[0]]
        itg = do_integration(expr, solvars, phys, mesh, kind, attrs, order, num)

        return itg



'''

   Complex Coefficient

    Handle Complex Coefficint as a single coefficint
    These classes are derived from RealImagCoefficientGen.

'''
import abc
from abc import ABC, abstractmethod

from numpy.linalg import inv, det

from petram.mfem_config import use_parallel
if use_parallel:
    import mfem.par as mfem
else:
    import mfem.ser as mfem


from petram.phys.phys_model import (PhysCoefficient,
                                    VectorPhysCoefficient,
                                    MatrixPhysCoefficient,
                                    PhysConstant,
                                    PhysVectorConstant,
                                    PhysMatrixConstant,
                                    Coefficient_Evaluator)

import petram.debug
dprint1, dprint2, dprint3 = petram.debug.init_dprints('PyComplexCoefficient')


'''
  Real/Imag Coefficient (mfem coefficients)

'''


class PyRealCoefficient(mfem.PyCoefficient):
    def __init__(self, coeff):
        self.coeff = coeff
        mfem.PyCoefficient.__init__(self)

    def Eval(self, T, ip):
        v = self.coeff.eval(T, ip)
        return v.real


class PyImagCoefficient(mfem.PyCoefficient):
    def __init__(self, coeff):
        self.coeff = coeff
        mfem.PyCoefficient.__init__(self)

    def Eval(self, T, ip):
        v = self.coeff.eval(T, ip)
        return v.imag


class PyRealVectorCoefficient(mfem.VectorPyCoefficient):
    def __init__(self, coeff):
        self.coeff = coeff
        mfem.VectorPyCoefficient.__init__(self, coeff.vdim)

    def Eval(self, K, T, ip):
        M = self.coeff.eval(T, ip)
        K.SetSize(M.shape[0])
        return K.Assign(M.real)


class PyImagVectorCoefficient(mfem.VectorPyCoefficient):
    def __init__(self, coeff):
        self.coeff = coeff
        mfem.VectorPyCoefficient.__init__(self, coeff.vdim)

    def Eval(self, K, T, ip):
        M = self.coeff.eval(T, ip)
        K.SetSize(M.shape[0])
        return K.Assign(M.imag)


class PyRealMatrixCoefficient(mfem.MatrixPyCoefficient):
    def __init__(self, coeff):
        self.coeff = coeff
        assert coeff.height == coeff.width, "not supported"
        mfem.MatrixPyCoefficient.__init__(self, coeff.height)

    def Eval(self, K, T, ip):
        M = self.coeff.eval(T, ip)
        K.SetSize(M.shape[0], M.shape[1])
        return K.Assign(M.real)


class PyImagMatrixCoefficient(mfem.MatrixPyCoefficient):
    def __init__(self, coeff):
        self.coeff = coeff
        assert coeff.height == coeff.width, "not supported"
        mfem.MatrixPyCoefficient.__init__(self, coeff.height)

    def Eval(self, K, T, ip):
        M = self.coeff.eval(T, ip)
        K.SetSize(M.shape[0], M.shape[1])
        return K.Assign(M.imag)


'''

  Complex Coefficient.

'''


class RealImagCoefficientGen(ABC):
    # abstract
    def __init__(self, kind, vdim=None, width=None, height=None):
        self._kind = kind
        if vdim is not None:
            self._vdim = vdim
        if width is not None:
            self._width = width
        if height is not None:
            self._height = height

    @abstractmethod
    def get_real_coefficient(self):
        pass

    @abstractmethod
    def get_imag_coefficient(self):
        pass

    @abstractmethod
    def get_realimag_coefficient(self, real):
        pass

    @abstractmethod
    def eval(self, T, ip):
        pass

    @property
    def kind(self):
        return self._kind

    @property
    def width(self):
        return self._width

    @property
    def height(self):
        return self._height

    @property
    def vdim(self):
        return self._vdim

    def is_matrix(self):
        return self.kind == 'matrix'

    def is_vector(self):
        return self.kind == 'vector'

    def __pow__(self, exponent):
        assert self.kind == 'scalar', "Pow is defined only for scalar"

        class PyComplexPowCoefficient(CC_Scalar):
            def __init__(self, coeff, pow):
                self.pow = pow
                CC_Scalar.__init__(self)
                self.coeff = coeff

            def eval(self, T, ip):
                v = self.coeff.eval(T, ip)
                v = (v)**(self.pow)
                return v
        return PyComplexPowCoefficient(self, exponent)

    def __add__(self, other):
        assert self.kind == self.kind, "__add__ must add the same coefficient types"

        class PyComplexSumSCoefficient(CC_Scalar):
            def __init__(self, coeff1, coeff2):
                CC_Scalar.__init__(self)
                self.coeff = coeff1
                self.coeff2 = coeff2

            def eval(self, T, ip):
                v1 = self.coeff.eval(T, ip)
                v2 = self.coeff2.eval(T, ip)
                return v1 + v2

        class PyComplexSumVCoefficient(CC_Vector):
            def __init__(self, coeff1, coeff2):
                CC_Vector.__init__(self, coeff1.vdim)
                self.coeff = coeff1
                self.coeff2 = coeff2

            def eval(self, T, ip):
                v1 = self.coeff.eval(T, ip)
                v2 = self.coeff2.eval(T, ip)
                return v1 + v2

        class PyComplexSumMCoefficient(CC_Matrix):
            def __init__(self, coeff1, coeff2):
                self.coeff = coeff1
                self.coeff2 = coeff2
                CC_Matrix.__init__(self, coeff1.height, coeff1.width)

            def eval(self, T, ip):
                v1 = self.coeff.eval(T, ip)
                v2 = self.coeff2.eval(T, ip)
                return v1 + v2
        if self.kind == "scalar":
            return PyComplexSumSCoefficient(self, other)
        if self.kind == "vector":
            return PyComplexSumVCoefficient(self, other)
        if self.kind == "matrix":
            return PyComplexSumMCoefficient(self, other)

    def __radd__(self, other):
        from petram.mfem_config import numba_debug
        from petram.phys.numba_coefficient import NumbaCoefficient

        if isinstance(other, NumbaCoefficient):
            if numba_debug:
                dprint1(
                    "in __radd__: NumbaCoefficient is converted to PyComplexCoefficient")
            obj = complex_coefficient_from_real_and_imag(
                other.real, other.imag)
            return obj + self
        else:
            return NotImplemented

    def __mul__(self, scale):
        assert not isinstance(
            scale, RealImagCoefficientGen), "multiplication is not defined"

        class PyComplexProductXCoefficient(CC_Scalar):
            def __init__(self, coeff, scale=1.0):
                self.scale = scale
                self.coeff = coeff
                CC_Scalar.__init__(self)

            def eval(self, T, ip):
                v = self.coeff.eval(T, ip)
                v *= self.scale
                return v
        obj = PyComplexProductXCoefficient(self, scale)

    def __getitem__(self, arg):
        '''
           A{1]  --> arg is int
           A{[1, 2]]  --> arg is list
           A{[1, 2], [2,3]]  --> arg is tuple
        '''
        check = self.kind == 'matrix' or self.kind == 'vector'
        assert check, "slice is valid for vector and matrix"

        class PyComplexVectorSliceVectorCoefficient(CC_Vector):
            def __init__(self, coeff, slice1):
                CC_Vector.__init__(self, len(slice1))
                self.slice1 = slice1
                self.coeff = coeff

            def eval(self, T, ip):
                mat = self.coeff.eval(T, ip)
                return mat[self.slice1]

            @property
            def vdim(self):
                return self._vdim

        class PyComplexVectorSliceScalarCoefficient(CC_Scalar):
            def __init__(self, coeff, slice1):
                CC_Scalar.__init__(self)
                self.slice1 = slice1
                self.coeff = coeff

            def eval(self, T, ip):
                mat = self.coeff.eval(T, ip)
                return mat[self.slice1]

        class PyComplexMatrixSliceScalarCoefficient(CC_Scalar):
            def __init__(self, coeff, slice1, slice2):
                CC_Scalar.__init__(self)
                self.coeff = coeff
                self.slice1 = slice1
                self.slice2 = slice2

            def eval(self, T, ip):
                mat = self.coeff.eval(T, ip)
                return mat[self.slice1, self.slice2]

        class PyComplexMatrixSliceVectorCoefficient(CC_Vector):
            def __init__(self, coeff, slice1, slice2):
                self.slice1 = slice1
                self.slice2 = slice2
                if len(slice1) == 1 and len(slice2) > 1:
                    vdim = len(slice2)
                elif len(slice2) == 1 and len(slice1) > 1:
                    vdim = len(slice1)
                else:
                    assert False, "SliceVector output should be a vector" + \
                        str(slice1) + "/" + str(slice2)
                CC_Vector.__init__(self, vdim)
                self.coeff = coeff

            def eval(self, T, ip):
                mat = self.coeff.eval(T, ip)
                return mat[self.slice1, :][:, self.slice2].flatten()

            @property
            def vdim(self):
                return self._vdim

        class PyComplexMatrixSliceMatrixCoefficient(CC_Matrix):
            def __init__(self, coeff, slice1, slice2):

                self.slice1 = slice1
                self.slice2 = slice2
                CC_Matrix.__init__(self, len(slice1), len(slice2))
                self.coeff = coeff

            def eval(self, T, ip):
                mat = self.coeff.eval(T, ip)
                ret = mat[self.slice1, :][:, self.slice2]
                return ret

        if self.kind == "vector":
            slice1 = arg
            try:
                a = slice1[0]
            except:
                slice1 = (slice1, )

            if len(slice1) == 1:
                return PyComplexVectorSliceScalarCoefficient(self, slice1[0])
            elif len(slice1) > 1:
                return PyComplexVectorSliceVectorCoefficient(self, slice1)
            else:
                assert False, "slice size must be greater than 1"

        if self.kind == "matrix":
            slice1, slice2 = arg
            try:
                a = slice1[0]
            except:
                slice1 = (slice1, )
            try:
                a = slice2[0]
            except:
                slice2 = (slice2, )

            if len(slice1) == 1 and len(slice2) == 1:
                return PyComplexMatrixSliceScalarCoefficient(self, slice1[0], slice2[0])

            elif len(slice1) == 1 and len(slice2) > 1:
                return PyComplexMatrixSliceVectorCoefficient(self, slice1, slice2)

            elif len(slice2) == 1 and len(slice1) > 1:
                return PyComplexMatrixSliceVectorCoefficient(self, slice1, slice2)

            elif len(slice2) > 1 and len(slice1) > 1:
                return PyComplexMatrixSliceMatrixCoefficient(self, slice1, slice2)
            else:
                assert False, "slice size must be greater than 1"

    def inv(self):
        '''
        return inverse
        '''
        assert self.kind == 'matrix', "inverse is only valid for matrix"

        class PyComplexMatrixInvCoefficient(CC_Matrix):
            def __init__(self, coeff):
                CC_Matrix.__init__(self, coeff.height, coeff.width)
                self.coeff = coeff

            def eval(self, T, ip):
                mat = self.coeff.eval(T, ip)
                mat = inv(mat)
                return mat
        return PyComplexMatrixInvCoefficient(self)

    def adj(self):
        class PyComplexMatrixAdjCoefficient(CC_Matrix):
            def __init__(self, coeff):
                CC_Matrix.__init__(self, coeff.height, coeff.width)
                self.coeff = coeff

            def eval(self, T, ip):
                mat = self.coeff.eval(T, ip)
                determinant = det(mat)
                mat = inv(mat)
                return mat * determinant
        return PyComplexMatrixAdjCoefficient(self)

# constant coefficients


class PyComplexConstantBase(RealImagCoefficientGen):
    def get_realimag_coefficient(self, real):
        if real:
            return self.get_real_coefficient()
        else:
            return self.get_imag_coefficient()


class PyComplexConstant(PyComplexConstantBase):
    def __init__(self, value):
        self.value = value
        self._kind = "scalar"

    def eval(self, T, ip):
        return self.value

    def get_real_coefficient(self):
        return PhysConstant(self.value.real)

    def get_imag_coefficient(self):
        return PhysConstant(self.value.imag)

    def __pow__(self, exponent):
        return PyComplexConstant((self.value)**exponent)

    def __mul__(self, scale):
        return PyComplexConstant((self.value)*scale)

    def __add__(self, other):
        if isinstance(other, PyComplexConstant):
            return PyComplexConstant(self.value + other.value)
        return other.__add__(self)


class PyComplexVectorConstant(PyComplexConstantBase):
    def __init__(self, value):
        self._vdim = len(value)
        self._kind = "vector"

        vec1 = mfem.Vector(value.real.flatten())
        vec2 = mfem.Vector(value.imag.flatten())
        self.value = (vec1.GetDataArray() + 1j*vec2.GetDataArray()).flatten()

    def get_imag_coefficient(self):
        return PhysVectorConstant(self.value.imag)

    def eval(self, T, ip):
        return self.value

    def get_real_coefficient(self):
        return PhysVectorConstant(self.value.real)

    def get_imag_coefficient(self):
        return PhysVectorConstant(self.value.imag)

    def __mul__(self, scale):
        return PyComplexVectorConstant((self.value)*scale)

    def __add__(self, other):
        if isinstance(other, PyComplexVectorConstant):
            return PyComplexVectorConstant(self.value + other.value)
        return other.__add__(self)

    def __getitem__(self, slice1):
        vec = self.value
        try:
            _a = slice1[0]
        except:
            slice1 = (slice1, )

        if len(slice1) == 1:
            value = vec[slice1[0]]
            return PyComplexConstant(value)

        value = vec[slice1]
        return PyComplexVectorConstant(value.flatten())


class PyComplexMatrixConstant(RealImagCoefficientGen):
    def __init__(self, value):
        self._width = value.shape[1]
        self._height = value.shape[0]
        self._kind = "matrix"

        mat1 = mfem.DenseMatrix(value.real)
        mat2 = mfem.DenseMatrix(value.imag)
        self.value = mat1.GetDataArray() + 1j*mat2.GetDataArray()

    def eval(self, T, ip):
        return self.value

    def get_real_coefficient(self):
        return PhysMatrixConstant(self.value.real)

    def get_imag_coefficient(self):
        return PhysMatrixConstant(self.value.imag)

    def get_realimag_coefficient(self, real):
        if real:
            return self.get_real_coefficient()
        else:
            return self.get_imag_coefficient()

    def __mul__(self, scale):
        return PyComplexMatrixConstant((self.value)*scale)

    def __add__(self, other):
        if isinstance(other, PyComplexMatrixConstant):
            return PyComplexMatrixConstant(self.value + other.value)
        return other.__add__(self)

    def __getitem__(self, arg):
        mat = self.value

        slice1, slice2 = arg

        try:
            _a = slice1[0]
        except:
            slice1 = (slice1, )
        try:
            _a = slice2[0]
        except:
            slice2 = (slice2, )
        if len(slice1) == 1 and len(slice2) == 1:
            value = mat[slice1[0], slice2[0]]
            return PyComplexConstant(value)
        if len(slice1) > 1 and len(slice2) > 1:
            value = mat[slice1, :][:, slice2]
            return PyComplexMatrixConstant(value)

        value = mat[slice1, slice2]
        return PyComplexVectorConstant(value.flatten())

    def inv(self):
        mat = inv(self.value)
        return PyComplexMatrixConstant(mat)

    def adj(self):
        mat = self.value
        determinant = det(mat)
        mat = inv(mat)
        return PyComplexMatrixConstant(mat * determinant)


# ComplexCoefficientBase
#   when real and imaginary parts are given by know MFEM coefficient

class PyComplexCoefficientBase(RealImagCoefficientGen):
    def __init__(self, coeff1, coeff2):
        check1 = isinstance(coeff1,
                            (mfem.Coefficient,
                             mfem.VectorCoefficient,
                             mfem.MatrixCoefficient))
        check2 = isinstance(coeff2,
                            (mfem.Coefficient,
                             mfem.VectorCoefficient,
                             mfem.MatrixCoefficient)) or coeff2 is None

        assert check1, "coefficient must be mfem coefficient"
        assert check2, "coefficient must be mfem coefficient"

        self.coeff1 = coeff1
        self.coeff2 = coeff2

    def get_real_coefficient(self):
        return self.coeff1

    def get_imag_coefficient(self):
        return self.coeff2

    def get_realimag_coefficient(self, real):
        if real:
            return self.get_real_coefficient()
        else:
            return self.get_imag_coefficient()


class PyComplexCoefficient(PyComplexCoefficientBase):
    def __init__(self, coeff1, coeff2):
        PyComplexCoefficientBase.__init__(self, coeff1, coeff2)
        self._kind = "scalar"

    def eval(self, T, ip):
        if self.coeff1 is not None:
            vec = complex(self.coeff1.Eval(T, ip))
        else:
            vec = 0.0
        if self.coeff2 is not None:
            vec += 1j * self.coeff2.Eval(T, ip)
        return vec


class PyComplexVectorCoefficient(PyComplexCoefficientBase):
    def __init__(self, coeff1, coeff2):
        PyComplexCoefficientBase.__init__(self, coeff1, coeff2)
        self._kind = "vector"
        if self.coeff1 is not None:
            self._vdim = self.coeff1.GetVDim()
        elif self.coeff2 is not None:
            self._vdim = self.coeff1.GetVDim()
        else:
            assert False, "Either Real or Imag should be non zero"
        self._V = mfem.Vector(self._vdim)

    def eval(self, T, ip):
        if self.coeff1 is not None:
            self.coeff1.Eval(self._V, T, ip)
        else:
            self._V.Assign(0.0)
        mat = self._V.GetDataArray().astype(complex)
        if self.coeff2 is not None:
            self.coeff2.Eval(self._V, T, ip)
            mat += 1j * self._V.GetDataArray()
        return mat


class PyComplexMatrixCoefficient(PyComplexCoefficientBase):
    def __init__(self, coeff1, coeff2):
        PyComplexCoefficientBase.__init__(self, coeff1, coeff2)
        self._kind = "matrix"
        if self.coeff1 is not None:
            self._height = self.coeff1.GetHeight()
            self._width = self.coeff1.GetWidth()
        elif self.coeff2 is not None:
            self._height = self.coeff1.GetHeight()
            self._width = self.coeff1.GetWidth()
        else:
            assert False, "Either Real or Imag should be non zero"

        self._K = mfem.DenseMatrix(self._height, self._width)

    def eval(self, T, ip):
        if self.coeff1 is not None:
            self.coeff1.Eval(self._K, T, ip)
        else:
            self._K.Assign(0.0)
        mat = self._K.GetDataArray().astype(complex)
        if self.coeff2 is not None:
            self.coeff2.Eval(self._K, T, ip)
            mat += 1j * self._K.GetDataArray()
        return mat


def complex_coefficient_from_real_and_imag(coeffr, coeffi):
    if (isinstance(coeffr, mfem.MatrixCoefficient) or
            isinstance(coeffi, mfem.MatrixCoefficient)):
        return PyComplexMatrixCoefficient(coeffr, coeffi)

    elif (isinstance(coeffr, mfem.VectorCoefficient) or
          isinstance(coeffi, mfem.VectorCoefficient)):
        return PyComplexVectorCoefficient(coeffr, coeffi)

    elif (isinstance(coeffr, mfem.Coefficient) or
          isinstance(coeffi, mfem.Coefficient)):
        return PyComplexCoefficient(coeffr, coeffi)


# CCBase : for generic Python written complex coefficient
class CCBase(RealImagCoefficientGen):
    def __init__(self, kind, vdim=None, height=None, width=None):
        RealImagCoefficientGen.__init__(
            self, kind, vdim=vdim, height=height, width=width)

        from petram.debug import handle_allow_python_function_coefficient
        msg = "Python function coefficient is created (complex)"
        handle_allow_python_function_coefficient(msg)

    def get_real_coefficient(self):
        if self.kind == 'scalar':
            return PyRealCoefficient(self)
        if self.kind == 'vector':
            return PyRealVectorCoefficient(self)
        if self.kind == 'matrix':
            return PyRealMatrixCoefficient(self)
        assert False, "unsupported kind"
        return None

    def get_imag_coefficient(self):
        if self.kind == 'scalar':
            return PyImagCoefficient(self)
        if self.kind == 'vector':
            return PyImagVectorCoefficient(self)
        if self.kind == 'matrix':
            return PyImagMatrixCoefficient(self)
        assert False, "unsupported kind"
        return None

    def get_realimag_coefficient(self, real):
        if real:
            return self.get_real_coefficient()
        return self.get_imag_coefficient()


class CC_Scalar(CCBase):
    def __init__(self):
        CCBase.__init__(self, "scalar")


class CC_Vector(CCBase):
    def __init__(self, vdim):
        CCBase.__init__(self, "vector", vdim=vdim)


class CC_Matrix(CCBase):
    def __init__(self, height, width):
        CCBase.__init__(self, "matrix", height=height, width=width)



from petram.mfem_model import MFEM_ModelRoot
model.scripts.helpers.create_ns('global')
root = MFEM_ModelRoot()
model.param.setvar('mfem_model', root)
root['General'].new_ns('global')
#model.scripts.helpers.rebuild_ns()
import mfem.ser as mfem



'''
BlcokPreconditioner Model. 
'''
from petram.solver.mumps_model import MUMPSPreconditioner
from petram.mfem_config import use_parallel
import numpy as np

from petram.debug import flush_stdout
from petram.namespace_mixin import NS_mixin
from .solver_model import LinearSolverModel, LinearSolver

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('BlockSmoother')

if use_parallel:
    from petram.helper.mpi_recipes import *
    from mfem.common.parcsr_extra import *
    import mfem.par as mfem
    default_kind = 'hypre'

    from mpi4py import MPI
    num_proc = MPI.COMM_WORLD.size
    myid = MPI.COMM_WORLD.rank
    smyid = '{:0>6d}'.format(myid)
    from mfem.common.mpi_debug import nicePrint

else:
    import mfem.ser as mfem
    default_kind = 'scipy'


class BlockSmoother(LinearSolverModel, NS_mixin):
    hide_ns_menu = True
    has_2nd_panel = False
    accept_complex = False
    always_new_panel = False

    def __init__(self, *args, **kwargs):
        LinearSolverModel.__init__(self, *args, **kwargs)
        NS_mixin.__init__(self, *args, **kwargs)

    @classmethod
    def fancy_menu_name(self):
        return 'BlockPreconditioner'

    @classmethod
    def fancy_tree_name(self):
        return 'BlockPreconditioner'

    def does_linearsolver_choose_linearsystem_type(self):
        return False

    def supported_linear_system_type(self):
        return ["blk_interleave",
                "blk_merged_s",
                "blk_merged", ]

    def real_to_complex(self, solall, M=None):
        raise NotImplementedError(
            "bug. this method sould not be called")


class DiagonalPreconditioner(BlockSmoother):
    @classmethod
    def fancy_menu_name(self):
        return 'DiagonalPreconditioner'

    @classmethod
    def fancy_tree_name(self):
        return 'DiagonalPreconditioner'

    def panel1_param(self):
        import wx
        from petram.pi.widget_smoother import WidgetSmoother

        smp1 = [None, None, 99, {"UI": WidgetSmoother, "span": (1, 2)}]

        return [[None, [False, [''], [[], ]], 27, [{'text': 'advanced mode'},
                                                   {'elp': [
                                                       ['preconditioner', '', 0, None], ]},
                                                   {'elp': [smp1, ]}], ], ]

    def get_panel1_value(self):
        # this will set _mat_weight
        from petram.solver.solver_model import SolveStep
        p = self.parent
        while not isinstance(p, SolveStep):
            p = p.parent
            if p is None:
                assert False, "Solver is not under SolveStep"
        num_matrix = p.get_num_matrix(self.get_phys())

        all_dep_vars = self.root()['Phys'].all_dependent_vars(num_matrix,
                                                              self.get_phys(),
                                                              self.get_phys_range())

        prec = [x for x in self.preconditioners if x[0] in all_dep_vars]
        names = [x[0] for x in prec]
        for n in all_dep_vars:
            if not n in names:
                prec.append((n, ['None', 'None']))
        self.preconditioners = prec

        value = ((self.adv_mode, [self.adv_prc, ], [self.preconditioners, ]),)

        return value

    def import_panel1_value(self, v):
        self.preconditioners = v[0][2][0]
        self.adv_mode = v[0][0]
        self.adv_prc = v[0][1][0]

    def attribute_set(self, v):
        v = super(DiagonalPreconditioner, self).attribute_set(v)
        v['preconditioner'] = ''
        v['preconditioners'] = []
        v['adv_mode'] = False
        v['adv_prc'] = ''
        return v

    def get_possible_child(self):
        from petram.solver.mumps_model import MUMPSPreconditionerModel
        from petram.solver.krylov import KrylovModel, KrylovSmoother
        return KrylovSmoother, MUMPSPreconditionerModel

    def get_possible_child_menu(self):
        from petram.solver.mumps_model import MUMPSPreconditionerModel
        from petram.solver.krylov import KrylovModel, KrylovSmoother
        choice = [("Blocks", KrylovSmoother),
                  ("!", MUMPSPreconditionerModel), ]
        return choice

    def prepare_solver(self, opr, engine):
        def get_operator_block(r, c):
            # if linked_op exists (= op is set from python).
            # try to get it
            # print(self.opr._linked_op)
            if hasattr(opr, "_linked_op"):
                try:
                    return opr._linked_op[(r, c)]
                except KeyError:
                    return None
            else:
                blk = opr.GetBlock(r, c)
                if use_parallel:
                    return mfem.Opr2HypreParMat(blk)
                else:
                    return mfem.Opr2SparseMat(blk)

        names = engine.masked_dep_var_names()

        if self.adv_mode:
            expr = self.adv_prc
            gen = eval(expr, self._global_ns)
            gen.set_param(A, names, engine, self)
            M = gen()

        else:
            prcs_gui = dict(self.preconditioners)

            ls_type = self.get_solve_root().get_linearsystem_type_from_modeltree()
            phys_real = self.get_solve_root().is_allphys_real()

            if ls_type == 'blk_interleave' and not phys_real:
                names = sum([[n, n] for n in names], [])

            import petram.helper.preconditioners as prcs

            g = prcs.DiagonalPrcGen(
                opr=opr, engine=engine, gui=self, name=names)
            M = g()

            pc_block = {}

            for k, n in enumerate(names):
                prctxt = prcs_gui[n][1] if use_parallel else prcs_gui[n][0]
                if prctxt == "None":
                    continue
                if prctxt.find("(") == -1:
                    prctxt = prctxt + "()"
                prcargs = "(".join(prctxt.split("(")[-1:])

                nn = prctxt.split("(")[0]

                if not n in pc_block:
                    # make a new one
                    if not nn in self:
                        try:
                            blkgen = getattr(prcs, nn)
                        except BaseException:
                            if nn in self._global_ns:
                                blkgen = self._global_ns[nn]
                            else:
                                raise

                        blkgen.set_param(g, n)
                        blk = eval("blkgen(" + prcargs)
                    else:
                        rr = engine.masked_dep_var_offset(n)
                        cc = engine.masked_r_dep_var_offset(n)
                        A = get_operator_block(rr, cc)
                        blk = self[nn].prepare_solver(A, engine)
                    M.SetDiagonalBlock(k, blk)
                    pc_block[n] = blk
                else:
                    M.SetDiagonalBlock(k, pc_block[n])
        return M


class DiagonalSmoother(DiagonalPreconditioner):
    @classmethod
    def fancy_menu_name(self):
        return 'DiagonalSmoother'

    @classmethod
    def fancy_tree_name(self):
        return 'DiagonalSmoother'



from petram.mfem_config import use_parallel
if use_parallel:
    import mfem.par as mfem
else:
    import mfem.ser as mfem
    
import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('WF_Natrual')

from petram.phys.vtable import VtableElement, Vtable
from petram.model import Domain, Bdry, Edge, Point, Pair
from petram.phys.phys_model import Phys, PhysModule

data =  (('label1', VtableElement(None, 
                                  guilabel = 'Natural BC',
                                  default = "",
                                  tip = "this does not constrain the model" )),)

class WF_Natural(Bdry, Phys):
    has_essential = False
    nlterms = []
    can_timedpendent = False
    has_3rd_panel = True
    vt  = Vtable(data)

    def __init__(self, **kwargs):
        super(WF_Natural, self).__init__(**kwargs)

    def attribute_set(self, v):
        Bdry.attribute_set(self, v)
        Phys.attribute_set(self, v)
        v['sel_readonly'] = False
        v['sel_index'] = []
        return v

    def get_essential_idx(self, kfes):
        return []






import os
import numpy as np
from scipy.sparse import coo_matrix
from collections import OrderedDict, defaultdict
from petram.helper.global_named_list import GlobalNamedList

debug = False

import petram.debug
dprint1, dprint2, dprint3 = petram.debug.init_dprints('mesh utils')

def distribute_shared_entity(pmesh):
    '''
    distribute entitiy numbering in master (owner) process
    '''
    from mpi4py import MPI
    myid = MPI.COMM_WORLD.rank
    nprc = MPI.COMM_WORLD.size   
    comm  = MPI.COMM_WORLD
        
    from mfem.common.mpi_debug import nicePrint, niceCall        
    
    from petram.helper.mpi_recipes import allgather, allgather_vector, gather_vector    
    master_entry = []
    local_data = {}
    master_data = {}
    MFEM3 = hasattr(pmesh, 'GroupFace')
    offset_v = np.hstack([0, np.cumsum(allgather(pmesh.GetNV()))])    
    offset_e = np.hstack([0, np.cumsum(allgather(pmesh.GetNEdges()))])
    offset_f = np.hstack([0, np.cumsum(allgather(pmesh.GetNFaces()))])

    ng = pmesh.GetNGroups()

    from mfem.par import intp
    def GroupEdge(j, iv):
        edge = intp(); o = intp()
        pmesh.GroupEdge(j, iv, edge, o)
        return edge.value()
    def GroupFace(j, iv):
        face = intp(); o = intp()
        pmesh.GroupFace(j, iv, face, o)
        return face.value()
    def GroupTriangle(j, iv):
        face = intp(); o = intp()
        pmesh.GroupTriangle(j, iv, face, o)
        return face.value()
    def GroupQuadrilateral(j, iv):
        face = intp(); o = intp()
        pmesh.GroupQuadrilateral(j, iv, face, o)
        return face.value()
     
    for j in range(ng):
        if j == 0: continue
        nv = pmesh.GroupNVertices(j)
        sv = np.array([pmesh.GroupVertex(j, iv) for iv in range(nv)])
        ne = pmesh.GroupNEdges(j)
        se = np.array([GroupEdge(j, iv) for iv in range(ne)])
        
        if MFEM3:        
            nf = pmesh.GroupNFaces(j)
            sf = np.array([GroupFace(j, iv) for iv in range(nf)])
        else:
            nt = pmesh.GroupNTriangles(j)
            nq = pmesh.GroupNQuadrilaterals(j)
            sf = np.array([GroupTriangle(j, iv) for iv in range(nt)] +
                          [GroupQuadrilateral(j, iv) for iv in range(nq)])

        data = (sv + offset_v[myid],
                se + offset_e[myid],
                sf + offset_f[myid])
            
            
        local_data[(pmesh.gtopo.GetGroupMasterRank(j),
                    pmesh.gtopo.GetGroupMasterGroup(j))] = data

        if pmesh.gtopo.IAmMaster(j):
            master_entry.append((myid, j,))
            '''
            mv = sv + offset_v[myid]
            me = se + offset_e[myid]
            mf = sf + offset_f[myid]
            data = (mv, me, mf)
            '''
        else:
            data = None
        master_data[(pmesh.gtopo.GetGroupMasterRank(j),
                     pmesh.gtopo.GetGroupMasterGroup(j))] = data
        
    master_entry = comm.gather(master_entry)
    if myid == 0:
        master_entry = sum(master_entry, [])
    master_entry = comm.bcast(master_entry)
    for entry in master_entry:
        master_id = entry[0]
        if master_id == myid:
            data = master_data[entry]
        else:
            data = None
        data = comm.bcast(data, root=master_id)
        if entry in master_data:
            master_data[entry] = data
            
    return local_data, master_data


def bdr_loop(mesh):
    use_parallel = hasattr(mesh, "GroupNVertices")
    
    ## fill line/surface loop info
    doms = np.unique(mesh.GetAttributeArray()).astype(int, copy=False)
    attrs =  mesh.GetBdrAttributeArray()
    
    if use_parallel:
        from petram.helper.mpi_recipes import allgather_vector
        
        doms = allgather_vector(doms).astype(int, copy=False)
        doms = np.unique(doms)
        
    loop= {k:[] for k in doms}

    for ibdr in range(mesh.GetNBE()):
        idx = mesh.GetBdrElementEdgeIndex(ibdr)
        elem1, elem2 = mesh.GetFaceElements(idx)
        if elem1 != -1:
            i = mesh.GetAttribute(elem1)
            loop[i].append(attrs[ibdr])
        if elem2 != -1:
            i = mesh.GetAttribute(elem2)
            loop[i].append(attrs[ibdr])
    for k in loop:
        loop[k] = list(set(loop[k]))

    return loop
    
def find_edge_corner(mesh):
    '''
    For 3D geometry
      find line (boundary between two bdr_attribute) and
      corner of lines
    '''
    use_parallel = hasattr(mesh, "GroupNVertices")
    
    if use_parallel:
        from mpi4py import MPI
        myid = MPI.COMM_WORLD.rank
        nprc = MPI.COMM_WORLD.size   
        comm  = MPI.COMM_WORLD
        
        from mfem.common.mpi_debug import nicePrint, niceCall        
        from petram.helper.mpi_recipes import allgather, allgather_vector, gather_vector
        from petram.mesh.mesh_utils import distribute_shared_entity        
        if not hasattr(mesh, "shared_info"):
            mesh.shared_info = distribute_shared_entity(mesh)
    else:
        myid = 0
        nprc = 1
    ndim =  mesh.Dimension()
    sdim =  mesh.SpaceDimension()    
    ne = mesh.GetNEdges()
    assert ndim == 3, "find_edge_corner is for 3D mesh"

    # 3D mesh
    get_edges = mesh.GetBdrElementEdges
    get_attr  = mesh.GetBdrAttribute
    iattr= mesh.GetBdrAttributeArray()  # min of this array is 1
    nattr = 0 if iattr.size == 0 else np.max(iattr)        
    nb = mesh.GetNBE()
    if mesh.GetNBE() == 0 and nprc == 1:    
        return {}, {}, {}, {}
    

    if use_parallel:
        offset = np.hstack([0, np.cumsum(allgather(mesh.GetNEdges()))])
        offsetf = np.hstack([0, np.cumsum(allgather(mesh.GetNFaces()))])
        offsetv = np.hstack([0, np.cumsum(allgather(mesh.GetNV()))])
        myoffset = offset[myid]
        myoffsetf = offsetf[myid]
        myoffsetv = offsetv[myid]                
        nattr = max(allgather(nattr))
        ne = sum(allgather(mesh.GetNEdges()))
    else:
        myoffset  = np.array(0, dtype=int)
        myoffsetf = np.array(0, dtype=int)
        myoffsetv = np.array(0, dtype=int)

    edges = defaultdict(list)
    iedges = np.arange(nb, dtype=int)
    
    if use_parallel:
        # eliminate slave faces from consideration
        iface = np.array([mesh.GetBdrElementEdgeIndex(i) for i in iedges],
                          dtype=int)+myoffsetf
        mask = np.array([True]*len(iface), dtype=bool)
        ld, md = mesh.shared_info
        for key in ld.keys():
            mid, g_in_master = key
            if mid == myid: continue
            iii = np.in1d(iedges, ld[key][2], invert = True)
            mask = np.logical_and(mask, iii)
        iedges = iedges[mask]
        
    # nicePrint(len(iedges)) np 1,2,4 gives 900... ok

    for i in iedges:
        ie, io = get_edges(i)
        ie += myoffset
        iattr = get_attr(i)
        edges[iattr].extend(list(ie))

    if use_parallel:
        # collect edges using master edge number
        # and gather it to a node.
        edgesc = {}
        ld, md = mesh.shared_info        
        for j in range(1, nattr+1):
            if j in edges:
               data = np.array(edges[j], dtype=int)
               for key in ld.keys():
                   mid, g_in_master = key
                   if mid == myid: continue
                   for le, me in zip(ld[key][1], md[key][1]):
                       iii =  np.where(data == le)[0]
                       data[iii] = me
            else:
                data = np.atleast_1d([]).astype(int)
            data = gather_vector(data, root = j % nprc)
            if data is not None: edgesc[j] = data
        edges = edgesc

    # for each iattr real edge appears only once
    for key in edges.keys():
        seen = defaultdict(int)
        for x in edges[key]: seen[x] +=1
        edges[key] = [k for k in seen if seen[k] == 1]
    
    #nicePrint('Num edges',
    nedge = sum([len(edges[k]) for k in edges])
    if nedge != 0:
        N = np.hstack([np.zeros(len(edges[k]), dtype=int)+k-1 for k in edges.keys()])
        M = np.hstack([np.array(edges[k]) for k in edges.keys()])
    else:
        N = np.atleast_1d([]).astype(int)
        M = np.atleast_1d([]).astype(int)        
    M = M.astype(int, copy = False)
    N = N.astype(int, copy = False)    

    if use_parallel:
        # send attribute to owner of edges
        for j in range(nprc):
            idx = np.logical_and(M >= offset[j], M < offset[j+1])
            Mpart = M[idx]
            Npart = N[idx]
            Mpart = gather_vector(Mpart, root = j)
            Npart = gather_vector(Npart, root = j)
            if j==myid: M2, N2 = Mpart, Npart
        M, N = M2, N2
        
    #nicePrint('unique edge', len(np.unique(M)))
    #nicePrint('N', len(N))    
    data = M*0+1
    table1 = coo_matrix((data, (M, N)), shape = (ne, nattr), dtype = int)

    csr = table1.tocsr()

    #embeded surface only touches to one iattr
    idx = np.where(np.diff(csr.indptr) >= 1)[0]
    csr = csr[idx, :]    

    # this is true bdr edges.
    bb_edges = defaultdict(list)
    indptr = csr.indptr; indices = csr.indices
    
    for i in range(csr.shape[0]):
        idxs = tuple(sorted(indices[indptr[i]:indptr[i+1]]+1))
        bb_edges[idxs].append(idx[i])
    bb_edges.default_factory = None

    # sort keys (= attribute set) 
    keys = list(bb_edges)
    if use_parallel:
        keys = comm.gather(keys)
        if myid == 0: keys = sum(keys, [])        
    sorted_key = None
    if myid == 0:
        sorted_key = list(set(keys))
        sorted_key.sort(key = lambda x:(len(x), x))

    if use_parallel: sorted_key = comm.bcast(sorted_key, root=0)
    bb_edgess = OrderedDict()
    for k in sorted_key:
        if k in bb_edges:
            bb_edgess[k] = bb_edges[k]
        else:
            bb_edgess[k] = []  # in parallel, put empty so that key order is kept
    bb_edges = bb_edgess

    '''
    res = []
    for key in sorted_key:
        tmp = allgather(len(bb_edges[key]))
        if myid == 0:
            res.append((key, sum(tmp)))
    if myid == 0: print res
    '''
    # at this point each node has its own edges populated in bb_edges (no shadow)
    ivert = {}

    for k in sorted_key:
        if len(bb_edges[k])>0:
            ivert[k] = np.hstack([mesh.GetEdgeVertices(i-myoffset)+ myoffsetv
                              for i in np.unique(bb_edges[k])]).astype(int)
        else:
            ivert[k] = np.atleast_1d([]).astype(int)

    if use_parallel:
        # convert shadow vertex to real
        for k in sorted_key:
            data = ivert[k]
            for key in ld:
                if key[0] == myid: continue
                for le, me in zip(ld[key][0], md[key][0]):
                   iii =  np.where(data == le)[0]
                   data[iii] = me
            ivert[k] = data
        ivertc = {}
        for j, k in enumerate(sorted_key):
            data = gather_vector(ivert[k], root = j % nprc)
            if data is not None:
                ivertc[k] = data
        ivert = ivertc

    corners = {}
    for key in ivert:
        seen = defaultdict(int)
        for iiv in ivert[key]:
            seen[iiv] += 1
        corners[key] = [kk for kk in seen if seen[kk]==1]

    if len(corners) == 0:
       u = np.atleast_1d([]).astype(int)
    else:
       u = np.unique(np.hstack([corners[key]
                  for key in corners])).astype(int, copy=False)

    # collect vertex on each node and gather to node 0
    u_own = u
    if use_parallel:
         u = np.unique(allgather_vector(u))
         u_own = u.copy()
         for key in ld:
             if key[0] == myid: continue
             for lv, mv in zip(ld[key][0], md[key][0]):
                 iii =  np.where(u == mv)[0]
                 u[iii] = lv
         idx = np.logical_and(u >= offsetv[myid], u < offsetv[myid+1])         
         u= u[idx]  # u include shared vertex
         idx = np.logical_and(u_own >= offsetv[myid], u_own < offsetv[myid+1])
         u_own = u_own[idx] # u_own is only owned vertex
         
    #nicePrint('u_own',mesh.GetNV(),",",  u_own)
    if len(u_own) > 0:
        vtx = np.vstack([mesh.GetVertexArray(i - myoffsetv) for i in u_own])
    else:
        vtx = np.atleast_1d([]).reshape(-1, sdim)
    if use_parallel:
        u_own = gather_vector(u_own)
        vtx   = gather_vector(vtx.flatten())

    # sort vertex  
    if myid == 0:
        vtx = vtx.reshape(-1, sdim)
        #print('vtx shape', vtx.shape)
        tmp = sorted([(k, tuple(x)) for k, x in enumerate(vtx)], key=lambda x:x[1])
        if len(tmp) > 0:
            vtx = np.vstack([x[1] for x in tmp])
            u_own = np.hstack([[u_own[x[0]] for x in tmp]]).astype(int)
            ivert=np.arange(len(vtx), dtype=int)+1
        else:
            vtx = np.atleast_1d([]).astype(float)
            u_own = np.atleast_1d([]).astype(int)
            u_own = np.atleast_1d([]).astype(int)
    if use_parallel:
        #if myid != 0:
        #    u_own = None; vtx = None
        u_own = comm.bcast(u_own)
        ivert=np.arange(len(u_own), dtype=int)+1
        for key in ld:
            if key[0] == myid: continue
            for lv, mv in zip(ld[key][0], md[key][0]):
                iii =  np.where(u_own == mv)[0]
                u_own[iii] = lv
        idx = np.logical_and(u_own >= offsetv[myid], u_own < offsetv[myid+1])
        u_own = u_own[idx]
        ivert = ivert[idx]
        #vtx = comm.bcast(vtx)
        #vtx = comm.bcast(vtx)[idx.flatten()]


    vert2vert = {iv: iu-myoffsetv for iv, iu in zip(ivert, u_own)}
    #nicePrint('vert2vert', vert2vert)

    vv_values = np.array([x[1] for x in vert2vert.items()])
    vv_keys = np.array([x[0] for x in vert2vert.items()])
    
    # mapping line index to vertex index (not MFFEM vertex id)
    line2vert = {}

    #nicePrint(corners)
    for j, key in enumerate(sorted_key):
        data = corners[key] if key in corners else None
        if use_parallel:
            data = comm.bcast(data, root = j % nprc)
            data = np.array(data, dtype=int)            
            for key2 in ld:
                if key2[0] == myid: continue
                for lv, mv in zip(ld[key2][0], md[key2][0]):
                    iii =  np.where(data == mv)[0]
                    data[iii] = lv
            idx = np.logical_and(data >= offsetv[myid],
                                 data < offsetv[myid+1])
            data = data[idx]
        else:
            data = np.array(data, dtype=int)                        
        data = list(data - myoffsetv)

        line2vert[j+1] = list(vv_keys[np.in1d(vv_values, data)])

        '''
        (this was origial very slow)
        line2vert[j+1] = [k for k in vert2vert
                          if vert2vert[k] in data]
        '''

    # finish-up edge data
    if use_parallel:
        # distribute edges, convert (add) from master to local
        # number
        for attr_set in bb_edges:
            data = sum(allgather(bb_edges[attr_set]), [])
            data = np.array(data, dtype=int)
            for key in ld:
                if key[0] == myid: continue
                for le, me in zip(ld[key][1], md[key][1]):
                   iii =  np.where(data == me)[0]
                   data[iii] = le
            
            idx = np.logical_and(data >= offset[myid], data < offset[myid+1])
            data = data[idx]
            bb_edges[attr_set] = list(data - myoffset)

        attrs = list(edges)
        attrsa = np.unique(sum(allgather(attrs), []))

        for a in attrsa:
            if a in attrs:
                data = np.array(edges[a], dtype=int)
            else:
                data = np.atleast_1d([]).astype(int)
            data = allgather_vector(data)
            
            for key in ld:
                if key[0] == myid: continue                
                for le, me in zip(ld[key][1], md[key][1]):
                   iii =  np.where(data == me)[0]
                   data[iii] = le
            idx = np.logical_and(data >= offset[myid], data < offset[myid+1])
            data = data[idx]
            edges[a] = list(data - myoffset)

    line2edge = {}
    for k, attr_set in enumerate(sorted_key):
        if attr_set in bb_edges:
            line2edge[k+1] = bb_edges[attr_set]
        else:
            line2edge[k+1] = []

    '''
    # debug find true (non-shadow) edges
    line2edge_true = {}
    for k, attr_set in enumerate(sorted_key):
        if attr_set in bb_edges:
            data = np.array(bb_edges[attr_set], dtype=int)
            for key in ld:
                if key[0] == myid: continue                                
                iii = np.in1d(data+myoffset, ld[key][1], invert = True)
                data = data[iii]
            line2edge_true[k+1] = data
        else:
            line2edge_true[k+1] = []
    nicePrint([sum(allgather(len(line2edge_true[key]))) for key in line2edge])
    '''
    surf2line = {k+1:[] for k in range(nattr)}
    for k, attr_set in enumerate(sorted_key):
        for a in attr_set: surf2line[a].append(k+1)

    if debug:
        g = GlobalNamedList(line2vert)
        g.sharekeys()
        gg = g.gather(nprc, overwrite = False).unique()
        if myid==0: print("debug (gathered line2vert)", gg)
    
    return surf2line, line2vert, line2edge, vert2vert

def find_corner(mesh):
    '''
    For 2D geometry
      find line (boundary between two bdr_attribute) and
      corner of lines
    '''
    use_parallel = hasattr(mesh, "GroupNVertices")
    
    if use_parallel:
        from mpi4py import MPI
        myid = MPI.COMM_WORLD.rank
        nprc = MPI.COMM_WORLD.size   
        comm  = MPI.COMM_WORLD
        
        from mfem.common.mpi_debug import nicePrint, niceCall        
        from petram.helper.mpi_recipes import allgather, allgather_vector, gather_vector
        from petram.mesh.mesh_utils import distribute_shared_entity        
        if not hasattr(mesh, "shared_info"):
            mesh.shared_info = distribute_shared_entity(mesh)
    else:
        myid = 0
        nprc = 1
        
    ndim =  mesh.Dimension()
    sdim =  mesh.SpaceDimension()    
    ne = mesh.GetNEdges()
    assert ndim == 2, "find_edge_corner is for 3D mesh"

    get_edges = mesh.GetElementEdges
    get_attr  = mesh.GetAttribute
    iattr= mesh.GetAttributeArray()     # min of this array is 1
    nattr = 0 if iattr.size == 0 else np.max(iattr)            
    nb = mesh.GetNE()
    nbe = mesh.GetNBE()
    if use_parallel:
        nbe = sum(allgather(nbe))
    if nbe == 0:    
        return {}, {}, {}

    if use_parallel:
        offset = np.hstack([0, np.cumsum(allgather(mesh.GetNEdges()))])
        offsetf = np.hstack([0, np.cumsum(allgather(mesh.GetNFaces()))])
        offsetv = np.hstack([0, np.cumsum(allgather(mesh.GetNV()))])
        myoffset = offset[myid]
        myoffsetf = offsetf[myid]
        myoffsetv = offsetv[myid]                
        nattr = max(allgather(nattr))
        ne = sum(allgather(mesh.GetNEdges()))
    else:
        myoffset  = np.array(0, dtype=int)
        myoffsetf = np.array(0, dtype=int)
        myoffsetv = np.array(0, dtype=int)
        
    if mesh.GetNBE() == 0: # some parallel node may have zero boundary
        battrs =  []
        iedges = np.array([], dtype=int)
    else:
        battrs =  mesh.GetBdrAttributeArray()
        iedges = np.hstack([mesh.GetBdrElementEdgeIndex(ibdr) for ibdr in
                        range(mesh.GetNBE())]).astype(int, copy=False)
    
    line2edge = GlobalNamedList()
    line2edge.setlists(battrs, iedges)

    if use_parallel:
        ld, md = mesh.shared_info                
        iedges = iedges+myoffset

    if use_parallel:
        for key2 in ld:
            if key2[0] == myid: continue
            iii = np.in1d(iedges, ld[key2][1], invert = True)
            if len(iii) == 0: continue
            iedges = iedges[iii]
            battrs = battrs[iii]

    line2realedge = GlobalNamedList()            
    line2realedge.setlists(battrs, iedges)
    
    line2realvert = GlobalNamedList()
    for key in line2realedge:
        data = np.hstack([mesh.GetEdgeVertices(i-myoffset)+ myoffsetv
                           for i in line2realedge[key]])
        if use_parallel:        
            for key2 in ld:
                if key2[0] == myid: continue                
                for lv, mv in zip(ld[key2][0], md[key2][0]):
                   iii =  np.where(data == lv)[0]
                   data[iii] = mv
        line2realvert[key] = data

    line2realvert.sharekeys().gather(nprc, distribute=True)

    corners = GlobalNamedList()
    for key in line2realvert:
        seen = defaultdict(int)
        for iiv in line2realvert[key]:
            seen[iiv] += 1
        corners[key] = [kk for kk in seen if seen[kk]==1]

    sorted_key = corners.sharekeys().globalkeys
    
    corners.allgather()
    u_own = np.unique(np.hstack([corners[key] for key in corners]).astype(int, copy=False))

    if use_parallel:
        idx = np.logical_and(u_own >= offsetv[myid], u_own < offsetv[myid+1])
        u_own = u_own[idx]

    if len(u_own) > 0:
        vtx = np.hstack([mesh.GetVertexArray(i - myoffsetv) for i in u_own])
    else:
        vtx = np.atleast_1d([])
    if use_parallel:
        vtx = gather_vector(vtx)
        u_own = gather_vector(u_own)

    # sort vertex  
    if myid == 0:
        vtx = vtx.reshape(-1, sdim)
        tmp = sorted([(k, tuple(x)) for k, x in enumerate(vtx)], key=lambda x:x[1])
        if len(tmp) > 0:
            vtx = np.vstack([x[1] for x in tmp])
            u_own = np.hstack([[u_own[x[0]] for x in tmp]]).astype(int)
            ivert=np.arange(len(vtx), dtype=int)+1
        else:
            u_own = np.atleast_1d([]).astype(int)
            ivert = np.atleast_1d([]).astype(int)

    if use_parallel:
        #if myid != 0:
        #    u_own = None; vtx = None
        u_own = comm.bcast(u_own)
        ivert=np.arange(len(u_own), dtype=int)+1
        for key in ld:
            if key[0] == myid: continue
            for lv, mv in zip(ld[key][0], md[key][0]):
                iii =  np.where(u_own == mv)[0]
                u_own[iii] = lv
        idx = np.logical_and(u_own >= offsetv[myid], u_own < offsetv[myid+1])
        u_own = u_own[idx]
        vtx = comm.bcast(vtx)
        vtx = comm.bcast(vtx)[idx.flatten()]
        ivert = ivert[idx]                          

    vert2vert = {iv: iu-myoffsetv for iv, iu in zip(ivert, u_own)}
    
    # mapping line index to vertex index (not MFFEM vertex id)
    line2vert = {}
    #nicePrint(corners)
    corners.bcast(nprc, distributed = True)
    for j, key in enumerate(sorted_key):
        data = corners[key]
        if use_parallel:                     
            for key2 in ld:
                if key2[0] == myid: continue
                for lv, mv in zip(ld[key2][0], md[key2][0]):
                    iii =  np.where(data == mv)[0]
                    data[iii] = lv
            idx = np.logical_and(data >= offsetv[myid],
                                 data < offsetv[myid+1])
            data = data[idx]
        data = list(data - myoffsetv)
        line2vert[j+1] = [k for k in vert2vert
                          if vert2vert[k] in data]
   
    if debug:
        g = GlobalNamedList(line2vert)
        g.sharekeys()
        gg = g.gather(nprc, overwrite = False).unique()
        if myid==0: print(gg)
        for i in range(nprc):
            if use_parallel:comm.barrier()
            if myid == i:
                for k in vert2vert:
                    print(myid, k, mesh.GetVertexArray(vert2vert[k]))
            if use_parallel:comm.barrier()                  
    return line2vert, line2edge, vert2vert

def get_extended_connectivity(mesh):
    
    ndim = mesh.Dimension()
    if ndim == 3:
        v2s = bdr_loop(mesh)
        s2l, l2v, l2e, v2v = find_edge_corner(mesh)        

    elif ndim == 2:
        v2s = None
        s2l = bdr_loop(mesh)
        l2v, l2e, v2v = find_corner(mesh)
        
    else:
        lp = bdr_loop(mesh)        
        v2s = None; s2l = None
        attrs = mesh.GetAttributeArray()        
        l2e = {k:list(np.where(attrs==k)[0]) for k in lp.keys()}
        l2v = bdr_loop(mesh)
        v = np.unique(np.hstack([lp[k] for k in lp.keys()]))
        battrs = mesh.GetBdrAttributeArray()
        v2v = {battr:mesh.GetBdrElementVertices(k)[0]
               for k, battr in enumerate(battrs)}
        
    from mfem.common.mpi_debug import nicePrint, niceCall                
    #nicePrint('s2l', s2l)
    mesh.extended_connectivity = {}
    me = mesh.extended_connectivity
    me['vol2surf']  = v2s
    me['surf2line'] = s2l
    me['line2edge'] = l2e
    me['line2vert'] = l2v
    me['vert2vert'] = v2v

def get_reverse_connectivity(mesh):
    def reverse_dict(d):
        dd = defaultdict(list)        
        for k in d:
            for v in d[k]:
                dd[v].append(k)
        return dict(dd)
    
    me = mesh.extended_connectivity
    if me['vol2surf'] is not None:
        me['surf2vol'] = reverse_dict(me['vol2surf'])
    if me['surf2line'] is not None:
        me['line2surf'] = reverse_dict(me['surf2line'])
    if me['line2vert'] is not None:
        me['vert2line'] = reverse_dict(me['line2vert'])        

def vol2line(v2s, s2l):
    v2l =  {}
    for v in v2s:
        l = np.unique(np.hstack([s2l[s] for s in v2s[v]]))
        l = l.astype(int)
        v2l[v] = list(l)
    return v2l

def line2vol(v2l):
    l2v = defaultdict(list)
    for v in v2l:
        for l in v2l[v]: l2v[l].append(v)
    return dict(l2v)

def line2surf(s2l):
    l2s = defaultdict(list)
    for s in s2l:
        for l in s2l[s]: l2s[l].append(s)
    return dict(l2s)
    
def populate_plotdata(mesh, table, cells, cell_data):
    from petram.mesh.mesh_utils import  get_extended_connectivity
    if not hasattr(mesh, 'extended_connectivity'):
        get_extended_connectivity(mesh)

    ndim = mesh.Dimension()
    ec = mesh.extended_connectivity
   
    cell_data['line'] = {}
    cell_data['vertex'] = {}
    l2e = ec['line2edge']
    v2v = ec['vert2vert']

    kedge = []

    if mesh.Dimension() >= 2:
        method = mesh.GetEdgeVertices
    else:
        method = mesh.GetElementVertices
        
    if len(l2e) > 0:
        #kedge = np.array(sum([[key]*len(l2e[key]) for key in l2e], [])).astype(int)
        kedge = np.hstack([[key]*len(l2e[key]) for key in l2e]).astype(int, copy=False)
        iverts = np.vstack([method(ie)
                        for key in l2e for ie in l2e[key]])
    else:
        iverts = np.atleast_1d([]).astype(int)
    cells['line'] = table[iverts]
    cell_data['line']['physical'] = np.array(kedge)

    kvert = np.array([key for key in v2v]).astype(int, copy=False)
    iverts = np.array([v2v[key] for key in v2v]).astype(int, copy=False)

    cells['vertex'] = table[iverts]    
    cell_data['vertex']['physical'] = kvert
    if ndim == 3:
        l_s_loop = [ec['surf2line'], ec['vol2surf']]
    elif ndim == 2:
        l_s_loop = [ec['surf2line'], None]
    else:
        l_s_loop = [None, None]

    iedge2bb = None # is it used?
    return l_s_loop

#
#  FaceOf, EdgeOf, PointOf
#
#  Note: don't convert kwargs to mesh=None, since it provides are
#        safe-separation between args and kwargs.
#
def FaceOf(*args, **kwargs):
    mesh=kwargs.pop('mesh', None)
    vols = list(np.atleast_1d(args).flatten())
    ec = mesh.extended_connectivity['vol2surf']
    if ec is None: return []
    return list(set(sum([list(ec[k]) for k in vols],[])))

def EdgeOf(*args, **kwargs):
    mesh=kwargs.pop('mesh', None)    
    faces = list(np.atleast_1d(args).flatten())
    ec = mesh.extended_connectivity['surf2line']
    if ec is None: return []    
    return list(set(sum([list(ec[k]) for k in faces],[])))

def PointOf(*args, **kwargs):
    mesh=kwargs.pop('mesh', None)        
    edges = list(np.atleast_1d(args).flatten())
    ec = mesh.extended_connectivity['line2vert']
    if ec is None: return []    
    return list(set(sum([list(ec[k]) for k in edges],[])))

def NeighborFaceOf(*args, **kwargs):
    # return Neighboring Face
    mesh=kwargs.pop('mesh', None)
    faces = list(np.atleast_1d(args).flatten())
    s2l = mesh.extended_connectivity['surf2line']
    e1 = EdgeOf(faces, mesh = mesh)
    neighbor = [k for k in s2l if len(np.intersect1d(EdgeOf(k, mesh=mesh),e1))!=0]
    neighbor = [k for k in neighbor if not k in faces]
    return neighbor

def CommonEdgeOf(faces1, faces2, **kwargs):
    # common edge of face group 1 and 2
    mesh=kwargs.pop('mesh', None)
    f1 = EdgeOf(faces1, mesh=mesh)
    f2 = EdgeOf(faces2, mesh=mesh)
    return list(np.intersect1d(f1,f2))



from petram.solver.solver_model import SolverInstance
import os
import numpy as np

from petram.model import Model
from petram.solver.solver_model import Solver
import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('DistanceSolver')
rprint = debug.regular_print('DistanceSolver')


class DistanceSolver(Solver):
    can_delete = True
    has_2nd_panel = False

    def attribute_set(self, v):
        super(DistanceSolver, self).attribute_set(v)
        v["solver_type"] = "heat flow"
        v["p_lap_p"] = 20
        v["p_lap_iter"] = 50
        v["heat_flow_t"] = 1.
        v["heat_flow_diffuse_iter"] = 1
        v["log_level"] = 1
        return v

    def panel1_param(self):
        return [  # ["Initial value setting",   self.init_setting,  0, {},],
            ["physics model", self.phys_model, 0, {}, ],
            ["method", self.solver_type, 1, {
                "values": ["heat flow", "p-Laplacian"]}],
            ["diffusion t (heat flow)", self.heat_flow_t, 300, {}, ],
            ["diffusion iter. (heat flow)",
             self.heat_flow_diffuse_iter, 400, {}, ],
            ["max power (p_Laplacian)", self.p_lap_p, 400, {}, ],
            ["newton iter(p_Laplacian)", self.p_lap_iter, 400, {}, ],
            ["log_level(0-1)", self.log_level, 400, {}],
            [None, self.init_only, 3, {"text": "initialize solution only"}],
            [None,
             self.clear_wdir, 3, {"text": "clear working directory"}],
            [None,
             self.save_parmesh, 3, {"text": "save parallel mesh"}],
            [None,
             self.use_profiler, 3, {"text": "use profiler"}], ]

    def get_panel1_value(self):
        return (  # self.init_setting,
            self.phys_model,
            self.solver_type,
            self.heat_flow_t,
            self.heat_flow_diffuse_iter,
            self.p_lap_p,
            self.p_lap_iter,
            self.log_level,
            self.init_only,
            self.clear_wdir,
            self.save_parmesh,
            self.use_profiler,)

    def import_panel1_value(self, v):
        #self.init_setting = str(v[0])
        self.phys_model = str(v[0])
        self.solver_type = str(v[1])
        self.heat_flow_t = float(v[2])
        self.heat_flow_diffuse_iter = int(v[3])
        self.p_lap_p = int(v[4])
        self.p_lap_iter = int(v[5])
        self.log_level = int(v[6])
        self.init_only = v[7]
        self.clear_wdir = v[8]
        self.save_parmesh = v[9]
        self.use_profiler = v[10]

    def get_editor_menus(self):
        return []

    def get_possible_child(self):
        return []

    def allocate_solver_instance(self, engine):
        if self.clear_wdir:
            engine.remove_solfiles()

        instance = DistanceSolverInstance(self, engine)
        return instance

    def get_matrix_weight(self, timestep_config):  # , timestep_weight):
        if timestep_config[0]:
            return [1, 0, 0]
        else:
            return [0, 0, 0]

    def get_custom_init(self):
        from petram.init_model import CustomInitSetting

        phys = self.parent.get_phys()
        init = CustomInitSetting(phys, value=[1.0, ])
        return [init, ]

    @debug.use_profiler
    def run(self, engine, is_first=True, return_instance=False):
        dprint1("Entering run (is_first=", is_first, ")", self.fullpath())
        if self.clear_wdir:
            engine.remove_solfiles()

        instance = DistanceSolverInstance(self, engine)

        instance.set_blk_mask()
        if return_instance:
            return instance

        if self.init_only:
            pass
        else:
            instance.solve()

        instance.save_solution(ksol=0,
                               skip_mesh=False,
                               mesh_only=False,
                               save_parmesh=self.save_parmesh)
        #engine.sol = instance.sol
        # instance.save_probe()
        is_first = False

        dprint1(debug.format_memory_usage())
        return is_first

    def does_solver_choose_linearsystem_type(self):
        return True

    def get_linearsystem_type_from_solvermodel(self):
        return 'blk_interleave'


class DistanceSolverInstance(SolverInstance):
    def __init__(self, gui, engine):
        SolverInstance.__init__(self, gui, engine)
        self.assembled = False
        self.linearsolver = None

    @property
    def blocks(self):
        return self.engine.assembled_blocks

    def set_linearsolver_model(self):
        # use its own solver from MFEM
        pass

    def compute_A(self, M, B, X, mask_M, mask_B):
        '''
        M[0] x = B

        return A and isAnew
        '''
        assert False, "not supposed to come here"
        # return M[0], True

    def compute_rhs(self, M, B, X):
        assert False, "not supposed to come here"

    def solve(self, update_operator=True):
        engine = self.engine
        phys_target = self.get_phys()
        phys_range = self.get_phys_range()

        engine.access_idx = 0
        name0 = phys_target[0].dep_vars[0]
        r_x = engine.r_x[engine.r_ifes(name0)]
        if len(phys_target[0].dep_vars) > 1:
            name1 = phys_target[0].dep_vars[1]
            dr_x = engine.r_x[engine.r_ifes(name1)]
            do_vector = True
        else:
            do_vector = False

        import mfem.par as mfem

        pfes_s = r_x.ParFESpace()
        pmesh = pfes_s.GetParMesh()

        filt_gf = mfem.ParGridFunction(pfes_s)

        # run heat solver
        if self.gui.solver_type == "heat flow":
            t_param = self.gui.heat_flow_t
            dx = mfem.dist_solver.AvgElementSize(pmesh)

            ds = mfem.dist_solver.HeatDistanceSolver(t_param * dx * dx)
            ds.smooth_steps = 0
            ds.vis_glvis = False

            ls_coeff = mfem.GridFunctionCoefficient(r_x)
            filt_gf.ProjectCoefficient(ls_coeff)

            ls_filt_coeff = mfem.GridFunctionCoefficient(filt_gf)

            ds.ComputeScalarDistance(ls_filt_coeff, r_x)
            if do_vector:
                ds.ComputeVectorDistance(ls_filt_coeff, dr_x)
        else:
            p = self.gui.p_lap_p
            newton_iter = self.gui.p_lap_iter
            ds = mfem.dist_solver.PLapDistanceSolver(p, newton_iter)

            ls_coeff = mfem.GridFunctionCoefficient(r_x)
            filt_gf.ProjectCoefficient(ls_coeff)
            ls_filt_coeff = mfem.GridFunctionCoefficient(filt_gf)

        # To DO FIX THIS
        #print(ds.print_level)
        #print(dir(ds.print_level))#.SetPrintLevel(self.gui.log_level)
        #ds.print_level.FirstAndLast().Summary()
        ds.ComputeScalarDistance(ls_filt_coeff, r_x)
        if do_vector:
            ds.ComputeVectorDistance(ls_filt_coeff, dr_x)

        return True
        '''
        filter = mfem.dist_solver.PDEFilter(pmesh, 10 * dx)
        '''
        # run PLapSolver

    def save_solution(self, ksol=0, skip_mesh=False,
                      mesh_only=False, save_parmesh=False):

        engine = self.engine
        phys_target = self.get_phys()

        if mesh_only:
            engine.save_sol_to_file(phys_target,
                                    mesh_only=True,
                                    save_parmesh=save_parmesh)
        else:
            engine.save_sol_to_file(phys_target,
                                    skip_mesh=skip_mesh,
                                    mesh_only=False,
                                    save_parmesh=save_parmesh)
            engine.save_extra_to_file(None)
        #engine.is_initialzied = False



'''
  list up the contents of sol directory
   solr
   soli
   solmesh
   probe_
   checkpoint_
'''
import os
from os.path import expanduser
from collections import defaultdict

def gather_soldirinfo(path):
    path = expanduser(path)
    checkpoints = {}    
    for nn in os.listdir(path):
        if (nn.startswith('checkpoint.') and
            nn.endswith('.txt')):
              fid = open(os.path.join(path, nn))
              lines = [l.strip().split(":") for l in fid.readlines()]
              lines = [(int(l[0]), float(l[1])) for l in lines]
              fid.close()
              solvername = nn.split('.')[1]
              checkpoints[solvername] = dict(lines)

    cp = defaultdict(dict)   ### cp["SolveStep1_TimeStep1"] = (1.0, dirname)
    for nn in os.listdir(path):
        if (nn.startswith('checkpoint_') and os.path.isdir(os.path.join(path, nn))):
            solvername = '_'.join(nn.split('_')[1:-1])
            idx = int(nn.split('_')[-1])
            if len(checkpoints[solvername]) > idx:            
                cp[solvername][(idx, checkpoints[solvername][idx])] = nn
    cp.default_factory=None

    probes = defaultdict(list)
    for nn in os.listdir(path):
        if nn.startswith('probe_'):
            if nn.find('.') == -1:
               signal = '_'.join(nn.split('_')[1:]) 
            else:
                #if int(nn.split('.')[1]) != 0: continue
                signal = '_'.join(nn.split('.')[0].split('_')[1:])
            probes[signal].append(nn)

    # sort probe files using the process number
    for key in probes:
        if len(probes[key]) > 1:
             xxx = [(int(x.split('.')[1]), x) for x in	probes[key]]
             xxx = [x[1] for x in sorted(xxx)]
             probes[key] = xxx            

    probes = dict(probes)
    cases = []
    cases = [(int(nn[5:]), nn) for nn in os.listdir(path) if nn.startswith('case')]
    cases = [xx[1] for xx in sorted(cases)]

    soldirinfo = {'checkpoint': dict(cp),
                  'probes': dict(probes),
                  'cases': cases}
    return soldirinfo

def gather_soldirinfo_s(path):
    try:
        info = gather_soldirinfo(path)
        result = (True, info)
    except:
        import traceback
        result = (False, traceback.format_exc())
        
    import petram.helper.pickle_wrapper as pickle    
    import binascii
    
    data = binascii.b2a_hex(pickle.dumps(result))
    
    return data
